from typing import List, Union, Type
from .utilities import determine_age_band
from .medicare_model import MedicareModel
from .category import Category
from .beneficiary import MedicareBeneficiary


class MedicareModelV28(MedicareModel):
    """
    This class represents the V28 Community Model for Medicare. It inherits from the MedicareModel class.

    Methods:
        __init__: Initializes the MedicareModelV24 instance.

        Overwrites:
            _get_normalization_factor: Retrieves the normalization factor based on the model year.
            _age_sex_edits: Applies age and sex edits to diagnosis codes.
            _determine_disease_interactions: Determines disease interactions based on Category objects and beneficiary information.

        Included for clarity:
            _determine_payment_count_category: Determines the payment count category based on the number of categories provided.
            _determine_age_gender_category: Determines the demographic category based on age, gender, and population.
            _determine_demographic_interactions: Determines demographic interactions based on gender, disability status, and Medicaid enrollment.

        New:
            _age_sex_edit_1: Applies age and sex edit 1 to a diagnosis code.
            _age_sex_edit_2: Applies age and sex edit 2 to a diagnosis code.
            _age_sex_edit_3: Applies age and sex edit 3 to a diagnosis code.
    """

    def __init__(self, year: Union[int, None] = None):
        super().__init__("v28", year)
        self.normalization_factor = self._get_normalization_factor(self.model_year)

    def _get_normalization_factor(self, year: int) -> float:
        """
        CMS updates normalization factor each year to apply to the risk scores. See:
        https://www.commonwealthfund.org/publications/explainer/2024/mar/how-government-updates-payment-rates-medicare-advantage-plans

        This dictionary is updated each year to include the normalization factor from the final announcement.

        Returns:
            float: The normalization factor.
        """
        norm_factor_dict = {
            2024: 1.015,
            2025: 1.045,
        }
        try:
            normalization_factor = norm_factor_dict[year]
        except KeyError:
            normalization_factor = 1

        return normalization_factor

    def _apply_hierarchies(
        self, categories: List[Type[Category]]
    ) -> List[Type[Category]]:
        """
        Filters out categories falling into hierarchies per the model hierarchy_definition file.
        In V28 there is a "heart interaction patch" in the file V283T3M which is easily
        applied in the hierachy step, thus this method is overwritten here.

        Args:
            categories (List[Type[Category]]): List of category objects to process.

        Returns:
            List[Type[Category]]: List of category objects after filtering.

        Notes:
            For each category, the codes dropped are tracked and assigned to the
            attribute "dropped_categories" of the category object.
        """
        category_list = [category.category for category in categories]
        dropped_codes_total = []

        # Patch for V28 Heart Conditions
        if "HCC223" in category_list and not any(
            category in category_list
            for category in ["HCC221", "HCC222", "HCC224", "HCC225", "HCC226"]
        ):
            dropped_codes_total.append("HCC223")

        for category in categories:
            dropped_codes = []
            if category.category in self.reference_files.hierarchy_definitions.keys():
                for remove_category in self.reference_files.hierarchy_definitions[
                    category.category
                ]["remove_code"]:
                    if remove_category in category_list:
                        dropped_codes.append(remove_category)
                        dropped_codes_total.append(remove_category)

            if dropped_codes:
                category.dropped_categories = dropped_codes

        # Remove objects from list
        final_categories = [
            category
            for category in categories
            if category.category not in dropped_codes_total
        ]

        return final_categories

    def _age_sex_edits(
        self, gender: str, age: int, diagnosis_code: str
    ) -> Union[List[str], None]:
        """
        Wrapper method to apply all model specific age and sex edits for a diagnosis code to
        category mapping. These are found in the model software file named something like
        "V28I0ED1".

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            age (int): Age of the individual.
            diagnosis_code (str): Diagnosis code to apply edits.

        Returns:
            Union[List[str], None]: List of categories after applying edits, or None if no edits applied.
        """
        new_category = self._age_sex_edit_1(gender, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_2(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_3(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_4(age, diagnosis_code)
        if new_category:
            return new_category

    def _age_sex_edit_1(self, gender: str, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 1 to a diagnosis code.

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if gender == "F" and dx_code in ["D66", "D67"]:
            return ["HCC112"]

    def _age_sex_edit_2(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 2 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if age < 18 and dx_code in [
            "J410",
            "J411",
            "J418",
            "J42",
            "J430",
            "J431",
            "J432",
            "J438",
            "J439",
            "J440",
            "J441",
            "J449",
            "J982",
            "J983",
        ]:
            return ["NA"]

    def _age_sex_edit_3(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 3 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if age < 50 and dx_code in [
            "C50011",
            "C50012",
            "C50019",
            "C50021",
            "C50022",
            "C50029",
            "C50111",
            "C50112",
            "C50119",
            "C50121",
            "C50122",
            "C50129",
            "C50211",
            "C50212",
            "C50219",
            "C50221",
            "C50222",
            "C50229",
            "C50311",
            "C50312",
            "C50319",
            "C50321",
            "C50322",
            "C50329",
            "C50411",
            "C50412",
            "C50419",
            "C50421",
            "C50422",
            "C50429",
            "C50511",
            "C50512",
            "C50519",
            "C50521",
            "C50522",
            "C50529",
            "C50611",
            "C50612",
            "C50619",
            "C50621",
            "C50622",
            "C50629",
            "C50811",
            "C50812",
            "C50819",
            "C50821",
            "C50822",
            "C50829",
            "C50911",
            "C50912",
            "C50919",
            "C50921",
            "C50922",
            "C50929",
        ]:
            return ["HCC22"]

    def _age_sex_edit_4(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 4 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if age >= 2 and dx_code in [
            "P040",
            "P041",
            "P0411",
            "P0412",
            "P0413",
            "P0414",
            "P0415",
            "P0416",
            "P0417",
            "P0418",
            "P0419",
            "P041A",
            "P042",
            "P043",
            "P0440",
            "P0441",
            "P0442",
            "P0449",
            "P045",
            "P046",
            "P048",
            "P0481",
            "P0489",
            "P049",
            "P270",
            "P271",
            "P278",
            "P279",
            "P930",
            "P938",
            "P961",
            "P962",
        ]:
            return ["NA"]

    def _determine_disease_interactions(
        self, categories: List[Type[Category]], beneficiary: Type[MedicareBeneficiary]
    ) -> List[Type[Category]]:
        """
        Determines disease interactions based on provided Category objects and beneficiary information.

        Args:
            categories (List[Type[Category]]): List of Category objects representing disease categories.
            beneficiary (Type[MedicareBeneficiary]): Instance of MedicareBeneficiary representing the beneficiary information.

        Returns:
            List[Type[Category]]: List of Category objects representing the disease interactions.
        """
        category_list = [
            category.category for category in categories if category.type == "disease"
        ]
        cancer_list = ["HCC17", "HCC18", "HCC19", "HCC20", "HCC21", "HCC22", "HCC23"]
        diabetes_list = ["HCC35", "HCC36", "HCC37", "HCC38"]
        card_resp_fail_list = ["HCC211", "HCC212", "HCC213"]
        hf_list = ["HCC221", "HCC222", "HCC223", "HCC224", "HCC225", "HCC226"]
        chr_lung_list = ["HCC276", "HCC277", "HCC278", "HCC279", "HCC280"]
        kidney_v28_list = ["HCC326", "HCC327", "HCC328", "HCC329"]
        g_substance_use_disorder_v28_list = [
            "HCC135",
            "HCC136",
            "HCC137",
            "HCC138",
            "HCC139",
        ]
        g_pyshiatric_v28_list = ["HCC151", "HCC152", "HCC153", "HCC154", "HCC155"]
        neuro_v28_list = [
            "HCC180",
            "HCC181",
            "HCC182",
            "HCC190",
            "HCC191",
            "HCC192",
            "HCC195",
            "HCC196",
            "HCC198",
            "HCC199",
        ]
        ulcer_v28_list = ["HCC379", "HCC380", "HCC381", "HCC382"]

        cancer = any(category in category_list for category in cancer_list)
        diabetes = any(category in category_list for category in diabetes_list)
        card_resp_fail = any(
            category in category_list for category in card_resp_fail_list
        )
        hf = any(category in category_list for category in hf_list)
        chr_lung = any(category in category_list for category in chr_lung_list)
        kidney_v28 = any(category in category_list for category in kidney_v28_list)
        g_substance_use_disorder_v28 = any(
            category in category_list for category in g_substance_use_disorder_v28_list
        )
        g_pyshiatric_v28 = any(
            category in category_list for category in g_pyshiatric_v28_list
        )
        neuro_v28 = any(category in category_list for category in neuro_v28_list)
        ulcer_v28 = any(category in category_list for category in ulcer_v28_list)
        hcc238 = "HCC238" in category_list

        interactions_dict = {
            "DIABETES_HF_V28": all([diabetes, hf]),
            "HF_CHR_LUNG_V28": all([hf, chr_lung]),
            "HF_KIDNEY_V28": all([hf, kidney_v28]),
            "CHR_LUNG_CARD_RESP_FAIL_V28": all([chr_lung, card_resp_fail]),
            "HF_HCC238_V28": all([hf, hcc238]),
            "gSubUseDisorder_gPsych_V28": all(
                [g_substance_use_disorder_v28, g_pyshiatric_v28]
            ),
            "DISABLED_CANCER_V28": all([beneficiary.disabled, cancer]),
            "DISABLED_NEURO_V28": all([beneficiary.disabled, neuro_v28]),
            "DISABLED_HF_V28": all([beneficiary.disabled, hf]),
            "DISABLED_CHR_LUNG_V28": all([beneficiary.disabled, chr_lung]),
            "DISABLED_ULCER_V28": all([beneficiary.disabled, ulcer_v28]),
        }
        interaction_list = [key for key, value in interactions_dict.items() if value]

        category_count = self._determine_payment_count_category(category_list)
        if category_count:
            interaction_list.append(category_count)
        interactions = [
            Category(self.reference_files, beneficiary.risk_model_population, category)
            for category in interaction_list
        ]
        interactions.extend(categories)

        return interactions

    def _determine_payment_count_category(self, categories: list) -> str:
        """
        Determines the payment count category based on the number of categories provided.

        Args:
            categories (list): List of categories.

        Returns:
            str: Payment count category.
        """
        category_count = len(categories)
        category = None
        if category_count > 9:
            category = "D10P"
        elif category_count > 0:
            category = f"D{category_count}"

        return category

    def _determine_age_gender_category(
        self, age: int, gender: str, population: str
    ) -> str:
        """
        Determines the demographic category based on age, gender, and population.

        Args:
            age (int): Age of the individual.
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            population (str): Beneficiary model population used for scoring

        Returns:
            str: Demographic category based on age, gender, and population.
        """
        if population[:2] == "NE":
            demo_category_ranges = [
                "0_34",
                "35_44",
                "45_54",
                "55_59",
                "60_64",
                "65",
                "66",
                "67",
                "68",
                "69",
                "70_74",
                "75_79",
                "80_84",
                "85_89",
                "90_94",
                "95_GT",
            ]
        else:
            demo_category_ranges = [
                "0_34",
                "35_44",
                "45_54",
                "55_59",
                "60_64",
                "65_69",
                "70_74",
                "75_79",
                "80_84",
                "85_89",
                "90_94",
                "95_GT",
            ]

        demographic_category_range = determine_age_band(age, demo_category_ranges)

        if population[:2] == "NE":
            demographic_category = f"NE{gender}{demographic_category_range}"
        else:
            demographic_category = f"{gender}{demographic_category_range}"

        return demographic_category

    def _determine_demographic_interactions(
        self, gender: str, orig_disabled: bool, medicaid: bool
    ) -> List[str]:
        """
        Determines demographic interactions based on gender, disability status, and Medicaid enrollment.

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            orig_disabled (bool): Indicates if the individual was originally disabled.
            medicaid (bool): Indicates if the individual is enrolled in Medicaid.

        Returns:
            List[str]: List of demographic interaction labels.
        """
        demo_interactions = []
        if gender == "F" and orig_disabled:
            demo_interactions.append("OriginallyDisabled_Female")
        elif gender == "M" and orig_disabled:
            demo_interactions.append("OriginallyDisabled_Male")

        if medicaid:
            demo_interactions.append("LTIMCAID")

        return demo_interactions



import datetime
from typing import Union, Tuple


class Beneficiary:
    """
    Represents a beneficiary. As to why there is age and DOB: DOB is considered PHI.
    Thus to comply with HIPPA rules, it can be excluded and age used instead. However,
    one of the two is required.

    Attributes:
        gender (str): The gender of the beneficiary.
        age (int, optional): The age of the beneficiary.
        dob (str, optional): The date of birth of the beneficiary in ISO format.

    """

    def __init__(
        self, gender: str, age: Union[None, int] = None, dob: Union[None, str] = None
    ):
        """
        Initialize a Beneficiary object.

        Args:
            gender (str): The gender of the beneficiary.
            age (int, optional): The age of the beneficiary.
            dob (str, optional): The date of birth of the beneficiary in ISO format.

        """
        if age is None and dob is None:
            raise ValueError("Either age or dob must be provided.")
        self.gender = gender
        self.age = age
        self.dob = dob


class MedicareBeneficiary(Beneficiary):
    """
    Represents a Medicare beneficiary which expands upon the Beneficiary class and
    requires additional Medicare specific attributes: orec, medicaid, and population.
    See __init__ for more detailed description of these attributes

    Attributes:
        gender (str): The gender of the Medicare beneficiary.
        orec (str): The original reason for entitlement code.
        medicaid (bool): Indicates whether the beneficiary has Medicaid.
        population (str, optional): The Medicare population type (default is "CNA").
        age (int, optional): The age of the Medicare beneficiary.
        dob (str, optional): The date of birth of the Medicare beneficiary in ISO format.
        disabled (bool): Indicates if the beneficiary is disabled.
        orig_disabled (bool): Indiciates if the beneficiary was originally disabled.
        risk_model_age (int): Age of the benficiary used in the model scoring calculations.
                              Per CMS, it is age of the beneficiary as of February 1st of
                              the payment year.
        risk_model_population (str): The derived population for the beneficiary based on all
                                     beneficiary attributes. This is necessary as in the
                                     Community model, CMS New Enrollees are broken into four
                                     subpopulations based on Medicaid status and whether or
                                     not the beneficiary was "originally disabled". By only
                                     requiring "NE" to be passed in for a population value,
                                     users do not need to know how to determine the four
                                     additional subpopulations and the code does it for
                                     them. See _get_new_enrollee_population for more details.

    """

    def __init__(
        self,
        gender: str,
        orec: str,
        medicaid: bool,
        population="CNA",
        age: Union[None, int] = None,
        dob: Union[None, str] = None,
        model_year: Union[None, int] = None,
    ):
        """
        Initialize a MedicareBeneficiary object.

        Args:
            gender (str): The gender of the Medicare beneficiary.
            orec (str): The original reason entitlement code. See the below link for more information:
                        https://resdac.org/cms-data/variables/medicare-original-reason-entitlement-code-orec
            medicaid (bool): A boolean indicating whether the beneficiary has Medicaid.
            population (str, optional): The Medicare population type which the benficiary is
                                        associated with and the score is being computed for.
                                        Valid values are:
                                        CNA - Community, Non Dual, Aged (default)
                                        CND - Community, Non Dual, Disabled
                                        CPA - Community, Partial Dual, Aged
                                        CPD - Community, Partial Dual, Disabled
                                        CFA - Community, Full Dual, Aged
                                        CFD - Community, Full Dual, Disabled
                                        INS - Institutional
                                        NE - CMS New Enrollee
            age (int, optional): The age of the Medicare beneficiary.
            dob (str, optional): The date of birth of the Medicare beneficiary in ISO format.
            model_year (int, optional): The model year which this beneficiary object is associated with.
                              It is necessary to determine the age of the beneficiary if dob is passed in.
        """
        super().__init__(gender, age, dob)
        self.orec = orec
        self.medicaid = medicaid
        self.population = population
        self.model_year = model_year
        self.risk_model_age = self._determine_age(self.age, self.dob)
        self.disabled, self.orig_disabled = self._determine_disabled(
            self.age, self.orec
        )
        if self.population == "NE":
            self.risk_model_population = self._get_new_enrollee_population(
                self.risk_model_age, self.orec, self.medicaid
            )
        else:
            self.risk_model_population = population

    def _determine_age(self, age: int, dob: str) -> int:
        """
        Determine the age of the beneficiary based on either age or date of birth (DOB).

        This function addresses two design considerations:
        1. Date of birth (DOB) is considered Protected Health Information (PHI), thus
           allowing flexibility in handling PHI by accepting either age or DOB.
        2. The CMS Risk Adjustment Model uses age as of February 1st of the payment year.
           If DOB is provided, age needs to be computed relative to that date. That payment year
           must also be provided.

        Args:
            age (int): The age of the beneficiary.
            dob (str): The date of birth of the beneficiary in ISO format.

        Returns:
            int: The age of the beneficiary as of February 1st of the payment year.

        If age is provided, it is assumed to be correct as of February 1st of the payment year.
        If DOB is provided, it computes the age relative to February 1st of the payment year.
        """
        if dob:
            if self.model_year is None:
                raise ValueError(
                    "When date of birth is provided, model year must also be provided"
                )
            reference_date = datetime.datetime.fromisoformat(f"{self.model_year}-02-01")
            dt_dob = datetime.datetime.fromisoformat(dob)
            age = (
                reference_date.year
                - dt_dob.year
                - (
                    (reference_date.month, reference_date.day)
                    < (dt_dob.month, dt_dob.day)
                )
            )
        elif age:
            age = age

        return age

    def _determine_disabled(self, age: int, orec: str) -> Tuple[bool, bool]:
        """
        Determine disability status and original disability status based on age and
        original entitlement reason code.

        Args:
            age (int): The age of the individual.
            orec (str): The original reason for entitlement code.

        Returns:
            tuple: A tuple containing two boolean elements:
                - A bool indicating if the individual is disabled (True if disabled, False otherwise).
                - A bool indicating the original disability status (True if originally disabled, False otherwise).
        """
        if age < 65 and orec != "0":
            disabled = True
        else:
            disabled = False

        if orec in ("1", "3") and disabled == 0:
            orig_disabled = True
        else:
            orig_disabled = False

        return disabled, orig_disabled

    def _get_new_enrollee_population(self, age: int, orec: str, medicaid: bool):
        """
        Compute the new enrollee population for the Community model based on Medicaid status
        and whether or not the beneficiary was originally disabled.

        Args:
            age (int): The age of the beneficiary.
            orec (str): The original reason for entitlement category.
            medicaid (bool): A boolean indicating whether the beneficiary has Medicaid.

        Returns:
            str: The new enrollee population type.

        New enrollee populations:
            - NMCAID_NORIGDIS: Non-Medicaid and not Originally Disabled
            - MCAID_NORIGDIS: Medicaid and not Originally Disabled
            - NMCAID_ORIGDIS: Non-Medicaid and Originally Disabled
            - MCAID_ORIGDIS: Medicaid and Originally Disabled
        """
        ne_population = None

        if age >= 65 and orec == "1":
            ne_originally_disabled = True
        else:
            ne_originally_disabled = False
        if not ne_originally_disabled and not medicaid:
            ne_population = "NE_NMCAID_NORIGDIS"
        if not ne_originally_disabled and medicaid:
            ne_population = "NE_MCAID_NORIGDIS"
        if ne_originally_disabled and not medicaid:
            ne_population = "NE_NMCAID_ORIGDIS"
        if ne_originally_disabled and medicaid:
            ne_population = "NE_MCAID_ORIGDIS"

        if ne_population is None:
            raise ValueError(
                "Population value NE passed in, but unable to determine corresponding risk model population for associated beneficiary attributes"
            )

        return ne_population


class CommercialBeneficiary(Beneficiary):
    """
    Represents a Commercial beneficiary which expands upon the Beneficiary class and
    requires additional Commercial-specific attributes: metal level, enrollment days,
    CSR indicator, and last enrollment date.

    Attributes:
        gender (str): The gender of the Commercial beneficiary.
        metal_level (str): The metal level of the beneficiary's insurance plan (e.g., Bronze, Silver).
        enrollment_days (int): The number of days the beneficiary has been enrolled.
        csr_indicator (int): The cost-sharing reduction indicator. Values are 1, 2, 3, 4
        age (int, optional): The age of the Commercial beneficiary.
        dob (str, optional): The date of birth of the Commercial beneficiary in ISO format.
        model_year (int, optional): The model year associated with the beneficiary.
        last_enrollment_date (str, optional): The last enrollment date of the beneficiary in ISO format.
        risk_model_age (int): The age of the beneficiary used in model scoring calculations.
        risk_model_age_group (str): The age group of the beneficiary (Infant, Child, Adult).
        risk_model_population (str): The derived population for the beneficiary.
        enrollment_months (int): The number of months the beneficiary has been enrolled.

    See __init__ for more detailed descriptions of these attributes.
    """

    def __init__(
        self,
        gender: str,
        metal_level: str = "Bronze",
        enrollment_days: int = 365,
        csr_indicator: int = 1,
        age: Union[None, int] = None,
        dob: Union[None, str] = None,
        model_year: Union[None, int] = None,
        last_enrollment_date: Union[None, str] = None,
    ):
        """
        Initialize a CommercialBeneficiary object.

        Args:
            gender (str): The gender of the Commercial beneficiary.
            metal_level (str): The metal level of the beneficiary's insurance plan (default is "Bronze").
            enrollment_days (int): The number of days the beneficiary has been enrolled (default is 365).
            csr_indicator (int): The cost-sharing reduction indicator (default is 1).
            age (int, optional): The age of the Commercial beneficiary.
            dob (str, optional): The date of birth of the Commercial beneficiary in ISO format.
            model_year (int, optional): The model year associated with the beneficiary.
            last_enrollment_date (str, optional): The last enrollment date of the beneficiary in ISO format.
        """
        super().__init__(gender, age, dob)
        self.metal_level = metal_level
        self.enrollment_days = enrollment_days
        self.csr_indicator = csr_indicator
        self.model_year = model_year
        self.last_enrollment_date = last_enrollment_date
        self.risk_model_age = self._determine_age(self.age, self.dob)
        self.risk_model_age_group = self._determine_age_group(self.risk_model_age)
        self.risk_model_population = self.metal_level
        self.enrollment_months = self._determine_enrollment_months(self.enrollment_days)

    def _determine_age(
        self, age: Union[None, int] = None, dob: Union[None, str] = None
    ) -> int:
        """
        Determine the age of the beneficiary based on either age or date of birth (DOB).

        This function addresses two design considerations:
        1. Date of birth (DOB) is considered Protected Health Information (PHI), thus
           allowing flexibility in handling PHI by accepting either age or DOB.
        2. The HHS Risk Adjustment Model uses age as of the last eligibility date of that
           beneficiary for that benefit year. If DOB is provided, age needs to be computed
           relative to that date. That benefit year must also be provided.

        Args:
            age (int, optional): The age of the beneficiary.
            dob (str, optional): The date of birth of the beneficiary in ISO format.

        Returns:
            int: The age of the beneficiary as of the last enrollment date of the benefit year.

        Raises:
            ValueError: If DOB is provided but the last enrollment date is not provided.
        """
        if dob:
            if self.last_enrollment_date is None:
                raise ValueError(
                    "When date of birth is provided, last enrollment date must also be provided"
                )
            reference_date = datetime.datetime.fromisoformat(self.last_enrollment_date)
            dt_dob = datetime.datetime.fromisoformat(dob)
            model_age = (
                reference_date.year
                - dt_dob.year
                - (
                    (reference_date.month, reference_date.day)
                    < (dt_dob.month, dt_dob.day)
                )
            )
        elif age is not None:
            model_age = age

        return model_age

    def _determine_age_group(self, age: int) -> str:
        """
        Determine the age group of the beneficiary based on their age.

        Args:
            age (int): The age of the beneficiary.

        Returns:
            str: The age group of the beneficiary. Possible values are "Infant", "Child", "Adult".
        """
        if age < 2:
            return "Infant"
        elif 2 <= age < 21:
            return "Child"
        else:
            return "Adult"

    def _determine_enrollment_months(self, enrollment_days: int) -> int:
        """
        Determine the enrollment duration in months based on the number of enrollment days.

        Args:
            enrollment_days (int): The number of days the beneficiary has been enrolled.

        Returns:
            int: The number of months the beneficiary has been enrolled.
        """
        if 1 <= enrollment_days <= 31:
            return 1
        elif 32 <= enrollment_days <= 62:
            return 2
        elif 63 <= enrollment_days <= 92:
            return 3
        elif 93 <= enrollment_days <= 123:
            return 4
        elif 124 <= enrollment_days <= 153:
            return 5
        elif 154 <= enrollment_days <= 184:
            return 6
        elif 185 <= enrollment_days <= 214:
            return 7
        elif 215 <= enrollment_days <= 245:
            return 8
        elif 246 <= enrollment_days <= 275:
            return 9
        elif 276 <= enrollment_days <= 306:
            return 10
        elif 307 <= enrollment_days <= 335:
            return 11
        elif 336 <= enrollment_days <= 366:
            return 12
        else:
            return 0



from typing import List, Union, Type
from .utilities import determine_age_band
from .commercial_model import CommercialModel
from .category import Category
from .beneficiary import CommercialBeneficiary


class CommercialModelV07(CommercialModel):
    """
    This class represents the V07 Model for Commercial. It inherits from the CommercialModel class.

    Methods:
        __init__: Initializes the CommercialModelV07 instance.

        Overwrites:
            _age_sex_edits: Applies age and sex edits to diagnosis codes.
            _determine_disease_interactions: Determines disease interactions based on Category objects and beneficiary information.

        Included for clarity:
            _determine_infant_severity_level
            _determine_infant_maturity_status
            _determine_age_gender_category
            _determine_demographic_interactions

        New:
            _age_sex_edit_1 to _age_sex_edit_16: Applies 16 different age and sex edits to a diagnosis code.
    """

    reference_files_version_dict = {
        2023: {
            "version": "3.0",
            "description": "April 10, 2024: Final 2023 Benefit Year Risk Adjustment HHS-Developed Risk Adjustment Model Algorithm 'Do It Yourself (DIY)' Software",
        },
        2024: {
            "version": "1.0",
            "description": "July 17, 2024: 2024 Benefit Year Risk Adjustment Updated HHS-Developed Risk Adjustment Model Algorithm 'Do It Yourself (DIY)' Software",
        },
        2025: {
            "version": "0.0",
            "description": "Using Benefit Year 2024 mappings; coefficients come from April 2, 2024 CMS-9895-F: Patient Protection and Affordable Care Act; HHS Notice of Benefit and Payment Parameters for 2025 (PDF)",
        },
    }

    def __init__(self, year: Union[int, None] = None):
        super().__init__("v07", year)

    def _age_sex_edits(
        self, gender: str, age: int, diagnosis_code: str
    ) -> Union[List[str], None]:
        """
        Wrapper method to apply all model specific age and sex edits for a diagnosis code to
        category mapping. These are found in the model software file named something like
        "V24I0ED1".

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            age (int): Age of the individual.
            diagnosis_code (str): Diagnosis code to apply edits.

        Returns:
            Union[List[str], None]: List of categories after applying edits, or None if no edits applied.
        """
        new_category = self._age_sex_edit_1(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_2(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_3(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_4(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_5(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_6(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_7(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_8(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_9(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_10(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_11(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_12(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_13(gender, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_14(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_15(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_16(age, diagnosis_code)
        if new_category:
            return new_category

    def _age_sex_edit_1(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 1 to a diagnosis code.

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if age < 18 and dx_code in [
            "C9100",
            "C9101",
            "C9102",
            "C9500",
            "C9501",
            "C9502",
            "C7400",
            "C7401",
            "C7402",
            "C7410",
            "C7411",
            "C7412",
            "C7490",
            "C7491",
            "C7492",
        ]:
            return ["HHS_HCC009"]

    def _age_sex_edit_2(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 2 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if age < 18 and dx_code in [
            "J410",
            "J411",
            "J418",
            "J42",
            "J440",
            "J441",
            "J4481",
            "J4489",
        ]:
            return ["HHS_HCC161_1"]

    def _age_sex_edit_3(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 3 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 2) and dx_code in [
            "K55011",
            "K55012",
            "K55019",
            "K55021",
            "K55022",
            "K55029",
            "K55031",
            "K55032",
            "K55039",
            "K55041",
            "K55042",
            "K55049",
            "K55051",
            "K55052",
            "K55059",
            "K55061",
            "K55062",
            "K55069",
            "K5530",
            "K5531",
            "K5532",
            "K5533",
        ]:
            return ["HHS_HCC042"]

    def _age_sex_edit_4(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 4 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 50) and dx_code in [
            "C50011",
            "C50012",
            "C50019",
            "C50021",
            "C50022",
            "C50029",
            "C50111",
            "C50112",
            "C50119",
            "C50121",
            "C50122",
            "C50129",
            "C50211",
            "C50212",
            "C50219",
            "C50221",
            "C50222",
            "C50229",
            "C50311",
            "C50312",
            "C50319",
            "C50321",
            "C50322",
            "C50329",
            "C50411",
            "C50412",
            "C50419",
            "C50421",
            "C50422",
            "C50429",
            "C50511",
            "C50512",
            "C50519",
            "C50521",
            "C50522",
            "C50529",
            "C50611",
            "C50612",
            "C50619",
            "C50621",
            "C50622",
            "C50629",
            "C50811",
            "C50812",
            "C50819",
            "C50821",
            "C50822",
            "C50829",
            "C50911",
            "C50912",
            "C50919",
            "C50921",
            "C50922",
            "C50929",
        ]:
            return ["HHS_HCC011"]

    def _age_sex_edit_5(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 5 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 2) and dx_code in [
            "J430",
            "J431",
            "J432",
            "J438",
            "J439",
            "J449",
            "J982",
            "J983",
            "F200",
            "F201",
            "F202",
            "F203",
            "F205",
            "F2081",
            "F2089",
            "F209",
            "F21",
            "F22",
            "F23",
            "F24",
            "F250",
            "F251",
            "F258",
            "F259",
            "F28",
            "F29",
            "F3010",
            "F3011",
            "F3012",
            "F3013",
            "F302",
            "F303",
            "F304",
            "F308",
            "F309",
            "F310",
            "F3110",
            "F3111",
            "F3112",
            "F3113",
            "F312",
            "F3130",
            "F3131",
            "F3132",
            "F314",
            "F315",
            "F3160",
            "F3161",
            "F3162",
            "F3163",
            "F3164",
            "F3170",
            "F3171",
            "F3172",
            "F3173",
            "F3174",
            "F3175",
            "F3176",
            "F3177",
            "F3178",
            "F3181",
            "F3189",
            "F319",
            "F322",
            "F323",
            "F332",
            "F333",
            "F440",
            "F441",
            "F4481",
            "F481",
            "F5000",
            "F5001",
            "F5002",
            "F502",
            "F600",
            "F601",
            "F602",
            "F603",
            "F604",
            "F605",
            "F606",
            "F607",
            "F6081",
            "F6089",
            "F609",
            "G47411",
            "G47419",
            "G47421",
            "G47429",
            "R4588",
            "T1491XA",
            "T360X2A",
            "T361X2A",
            "T362X2A",
            "T363X2A",
            "T364X2A",
            "T365X2A",
            "T366X2A",
            "T367X2A",
            "T368X2A",
            "T3692XA",
            "T370X2A",
            "T371X2A",
            "T372X2A",
            "T373X2A",
            "T374X2A",
            "T375X2A",
            "T378X2A",
            "T3792XA",
            "T380X2A",
            "T381X2A",
            "T382X2A",
            "T383X2A",
            "T384X2A",
            "T385X2A",
            "T386X2A",
            "T387X2A",
            "T38802A",
            "T38812A",
            "T38892A",
            "T38902A",
            "T38992A",
            "T39012A",
            "T39092A",
            "T391X2A",
            "T392X2A",
            "T39312A",
            "T39392A",
            "T394X2A",
            "T398X2A",
            "T3992XA",
            "T400X2A",
            "T401X2A",
            "T402X2A",
            "T403X2A",
            "T40412A",
            "T40422A",
            "T40492A",
            "T405X2A",
            "T40602A",
            "T40692A",
            "T40712A",
            "T40722A",
            "T408X2A",
            "T40902A",
            "T40992A",
            "T410X2A",
            "T411X2A",
            "T41202A",
            "T41292A",
            "T413X2A",
            "T4142XA",
            "T415X2A",
            "T420X2A",
            "T421X2A",
            "T422X2A",
            "T423X2A",
            "T424X2A",
            "T425X2A",
            "T426X2A",
            "T4272XA",
            "T428X2A",
            "T43012A",
            "T43022A",
            "T431X2A",
            "T43202A",
            "T43212A",
            "T43222A",
            "T43292A",
            "T433X2A",
            "T434X2A",
            "T43502A",
            "T43592A",
            "T43602A",
            "T43612A",
            "T43622A",
            "T43632A",
            "T43642A",
            "T43652A",
            "T43692A",
            "T438X2A",
            "T4392XA",
            "T440X2A",
            "T441X2A",
            "T442X2A",
            "T443X2A",
            "T444X2A",
            "T445X2A",
            "T446X2A",
            "T447X2A",
            "T448X2A",
            "T44902A",
            "T44992A",
            "T450X2A",
            "T451X2A",
            "T452X2A",
            "T453X2A",
            "T454X2A",
            "T45512A",
            "T45522A",
            "T45602A",
            "T45612A",
            "T45622A",
            "T45692A",
            "T457X2A",
            "T458X2A",
            "T4592XA",
            "T460X2A",
            "T461X2A",
            "T462X2A",
            "T463X2A",
            "T464X2A",
            "T465X2A",
            "T466X2A",
            "T467X2A",
            "T468X2A",
            "T46902A",
            "T46992A",
            "T470X2A",
            "T471X2A",
            "T472X2A",
            "T473X2A",
            "T474X2A",
            "T475X2A",
            "T476X2A",
            "T477X2A",
            "T478X2A",
            "T4792XA",
            "T480X2A",
            "T481X2A",
            "T48202A",
            "T48292A",
            "T483X2A",
            "T484X2A",
            "T485X2A",
            "T486X2A",
            "T48902A",
            "T48992A",
            "T490X2A",
            "T491X2A",
            "T492X2A",
            "T493X2A",
            "T494X2A",
            "T495X2A",
            "T496X2A",
            "T497X2A",
            "T498X2A",
            "T4992XA",
            "T500X2A",
            "T501X2A",
            "T502X2A",
            "T503X2A",
            "T504X2A",
            "T505X2A",
            "T506X2A",
            "T507X2A",
            "T508X2A",
            "T50902A",
            "T50912A",
            "T50992A",
            "T50A12A",
            "T50A22A",
            "T50A92A",
            "T50B12A",
            "T50B92A",
            "T50Z12A",
            "T50Z92A",
            "T510X2A",
            "T511X2A",
            "T512X2A",
            "T513X2A",
            "T518X2A",
            "T5192XA",
            "T520X2A",
            "T521X2A",
            "T522X2A",
            "T523X2A",
            "T524X2A",
            "T528X2A",
            "T5292XA",
            "T530X2A",
            "T531X2A",
            "T532X2A",
            "T533X2A",
            "T534X2A",
            "T535X2A",
            "T536X2A",
            "T537X2A",
            "T5392XA",
            "T540X2A",
            "T541X2A",
            "T542X2A",
            "T543X2A",
            "T5492XA",
            "T550X2A",
            "T551X2A",
            "T560X2A",
            "T561X2A",
            "T562X2A",
            "T563X2A",
            "T564X2A",
            "T565X2A",
            "T566X2A",
            "T567X2A",
            "T56812A",
            "T56822A",
            "T56892A",
            "T5692XA",
            "T570X2A",
            "T571X2A",
            "T572X2A",
            "T573X2A",
            "T578X2A",
            "T5792XA",
            "T5802XA",
            "T5812XA",
            "T582X2A",
            "T588X2A",
            "T5892XA",
            "T590X2A",
            "T591X2A",
            "T592X2A",
            "T593X2A",
            "T594X2A",
            "T595X2A",
            "T596X2A",
            "T597X2A",
            "T59812A",
            "T59892A",
            "T5992XA",
            "T600X2A",
            "T601X2A",
            "T602X2A",
            "T603X2A",
            "T604X2A",
            "T608X2A",
            "T6092XA",
            "T6102XA",
            "T6112XA",
            "T61772A",
            "T61782A",
            "T618X2A",
            "T6192XA",
            "T620X2A",
            "T621X2A",
            "T622X2A",
            "T628X2A",
            "T6292XA",
            "T63002A",
            "T63012A",
            "T63022A",
            "T63032A",
            "T63042A",
            "T63062A",
            "T63072A",
            "T63082A",
            "T63092A",
            "T63112A",
            "T63122A",
            "T63192A",
            "T632X2A",
            "T63302A",
            "T63312A",
            "T63322A",
            "T63332A",
            "T63392A",
            "T63412A",
            "T63422A",
            "T63432A",
            "T63442A",
            "T63452A",
            "T63462A",
            "T63482A",
            "T63512A",
            "T63592A",
            "T63612A",
            "T63622A",
            "T63632A",
            "T63692A",
            "T63712A",
            "T63792A",
            "T63812A",
            "T63822A",
            "T63832A",
            "T63892A",
            "T6392XA",
            "T6402XA",
            "T6482XA",
            "T650X2A",
            "T651X2A",
            "T65212A",
            "T65222A",
            "T65292A",
            "T653X2A",
            "T654X2A",
            "T655X2A",
            "T656X2A",
            "T65812A",
            "T65822A",
            "T65832A",
            "T65892A",
            "T6592XA",
            "T71112A",
            "T71122A",
            "T71132A",
            "T71152A",
            "T71162A",
            "T71192A",
            "T71222A",
            "T71232A",
            "X710XXA",
            "X711XXA",
            "X712XXA",
            "X713XXA",
            "X718XXA",
            "X719XXA",
            "X72XXXA",
            "X730XXA",
            "X731XXA",
            "X732XXA",
            "X738XXA",
            "X739XXA",
            "X7401XA",
            "X7402XA",
            "X7409XA",
            "X748XXA",
            "X749XXA",
            "X75XXXA",
            "X76XXXA",
            "X770XXA",
            "X771XXA",
            "X772XXA",
            "X773XXA",
            "X778XXA",
            "X779XXA",
            "X780XXA",
            "X781XXA",
            "X782XXA",
            "X788XXA",
            "X789XXA",
            "X79XXXA",
            "X80XXXA",
            "X810XXA",
            "X811XXA",
            "X818XXA",
            "X820XXA",
            "X821XXA",
            "X822XXA",
            "X828XXA",
            "X830XXA",
            "X831XXA",
            "X832XXA",
            "X838XXA",
        ]:
            return ["NA"]

    def _age_sex_edit_6(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 6 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 21) and dx_code in [
            "H353210",
            "H353211",
            "H353212",
            "H353213",
            "H353220",
            "H353221",
            "H353222",
            "H353223",
            "H353230",
            "H353231",
            "H353232",
            "H353233",
            "H353290",
            "H353291",
            "H353292",
            "H353293",
        ]:
            return ["NA"]

    def _age_sex_edit_7(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 7 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age > 0) and dx_code in [
            "P0500",
            "P0501",
            "P0502",
            "P0503",
            "P0504",
            "P0505",
            "P0506",
            "P0507",
            "P0508",
            "P0509",
            "P0510",
            "P0511",
            "P0512",
            "P0513",
            "P0514",
            "P0515",
            "P0516",
            "P0517",
            "P0518",
            "P0519",
            "P052",
            "P059",
            "P0700",
            "P0701",
            "P0702",
            "P0703",
            "P0710",
            "P0714",
            "P0715",
            "P0716",
            "P0717",
            "P0718",
            "P0720",
            "P0721",
            "P0722",
            "P0723",
            "P0724",
            "P0725",
            "P0726",
            "P0730",
            "P0731",
            "P0732",
            "P0733",
            "P0734",
            "P0735",
            "P0736",
            "P0737",
            "P0738",
            "P0739",
            "P080",
            "P081",
            "P0821",
            "P0822",
            "Z3800",
            "Z3801",
            "Z381",
            "Z382",
            "Z3830",
            "Z3831",
            "Z384",
            "Z385",
            "Z3861",
            "Z3862",
            "Z3863",
            "Z3864",
            "Z3865",
            "Z3866",
            "Z3868",
            "Z3869",
            "Z387",
            "Z388",
        ]:
            return ["NA"]

    def _age_sex_edit_8(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 8 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age >= 1) and dx_code == "Q894":
            return ["HHS_HCC097"]

    def _age_sex_edit_9(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 9 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age >= 2) and dx_code in [
            "K551",
            "K558",
            "K559",
            "P040",
            "P0411",
            "P0412",
            "P0413",
            "P0414",
            "P0415",
            "P0416",
            "P0417",
            "P0418",
            "P0419",
            "P041A",
            "P042",
            "P043",
            "P0440",
            "P0441",
            "P0442",
            "P0449",
            "P045",
            "P046",
            "P0481",
            "P0489",
            "P049",
            "P930",
            "P938",
            "P961",
            "P962",
            "Q390",
            "Q391",
            "Q392",
            "Q393",
            "Q394",
            "Q6410",
            "Q6411",
            "Q6412",
            "Q6419",
            "Q790",
            "Q791",
            "Q792",
            "Q793",
            "Q794",
            "Q7951",
        ]:
            return ["NA"]

    def _age_sex_edit_10(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 9 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age >= 2) and dx_code in ["P270", "P271", "P278", "P279"]:
            return ["HHS_HCC162"]

    def _age_sex_edit_11(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 11 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 6 or age > 18) and dx_code == "F3481":
            return ["NA"]

    def _age_sex_edit_12(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 11 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age >= 21) and dx_code in [
            "E700",
            "E701",
            "E7020",
            "E7021",
            "E7029",
            "E7030",
            "E70310",
            "E70311",
            "E70318",
            "E70319",
            "E70320",
            "E70321",
            "E70328",
            "E70329",
            "E70330",
            "E70331",
            "E70338",
            "E70339",
            "E7039",
            "E7040",
            "E7041",
            "E7049",
            "E705",
            "E7081",
            "E7089",
            "E709",
            "E710",
            "E71110",
            "E71111",
            "E71118",
            "E71120",
            "E71121",
            "E71128",
            "E7119",
            "E712",
            "E7130",
            "E71310",
            "E71311",
            "E71312",
            "E71313",
            "E71314",
            "E71318",
            "E7132",
            "E7139",
            "E7142",
            "E7150",
            "E71510",
            "E71511",
            "E71518",
            "E71520",
            "E71521",
            "E71522",
            "E71528",
            "E71529",
            "E7153",
            "E71540",
            "E71541",
            "E71542",
            "E71548",
            "E7200",
            "E7201",
            "E7202",
            "E7203",
            "E7204",
            "E7209",
            "E7210",
            "E7211",
            "E7212",
            "E7219",
            "E7220",
            "E7221",
            "E7222",
            "E7223",
            "E7229",
            "E723",
            "E724",
            "E7250",
            "E7251",
            "E7252",
            "E7253",
            "E7259",
            "E7281",
            "E7289",
            "E729",
            "E7420",
            "E7421",
            "E7429",
            "E744",
            "E74810",
            "E74818",
            "E74819",
            "E7489",
            "E749",
            "E771",
            "E8840",
            "E8841",
            "E8842",
            "E8843",
            "E8849",
            "H49811",
            "H49812",
            "H49813",
            "H49819",
        ]:
            return ["NA"]

    def _age_sex_edit_13(self, gender: str, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 13 to a diagnosis code.

        Args:
            gender (str): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if gender == "F" and dx_code in ["D66", "D67"]:
            return ["HHS_HCC075"]

    def _age_sex_edit_14(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 14 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 21) and dx_code in ["E10641", "E1011", "E1010"]:
            return ["HHS_HCC019"]

    def _age_sex_edit_15(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 15 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 21) and dx_code in [
            "E1022",
            "E1029",
            "E1021",
            "E1052",
            "E1059",
            "E1051",
            "E1041",
            "E1049",
            "E1044",
            "E1040",
            "E1043",
            "E1042",
            "E10610",
            "E10621",
            "E10620",
            "E10622",
            "E10618",
            "E10628",
            "E10638",
            "E10630",
            "E1069",
            "E103591",
            "E103412",
            "E103593",
            "E103559",
            "E103291",
            "E103552",
            "E103549",
            "E103393",
            "E103542",
            "E103539",
            "E10311",
            "E103219",
            "E103391",
            "E103532",
            "E103529",
            "E103531",
            "E103533",
            "E103392",
            "E103541",
            "E103543",
            "E103399",
            "E103551",
            "E10319",
            "E103553",
            "E103411",
            "E103292",
            "E103592",
            "E103413",
            "E103599",
            "E103523",
            "E103213",
            "E103319",
            "E103522",
            "E103521",
            "E103313",
            "E103519",
            "E103513",
            "E103212",
            "E103312",
            "E103512",
            "E103511",
            "E103311",
            "E103499",
            "E103211",
            "E103493",
            "E103492",
            "E1039",
            "E103299",
            "E103491",
            "E1037X9",
            "E1037X3",
            "E1037X2",
            "E1037X1",
            "E103293",
            "E103419",
            "E1036",
        ]:
            return ["HHS_HCC020"]

    def _age_sex_edit_16(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 16 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 21) and dx_code in ["E108", "E1065", "E10649", "E109"]:
            return ["HHS_HCC021"]

    def _determine_interactions(
        self,
        categories: List[Type[Category]],
        beneficiary: Type[CommercialBeneficiary],
        dropped_group_categories: List[Type[Category]],
    ) -> List[Type[Category]]:
        """
        Determines disease interactions based on provided Category objects and beneficiary information.
        This is relevant only for the Adult and Child groups

        Args:
            categories (List[Type[Category]]): List of Category objects representing disease categories.
            beneficiary (Type[CommercialBeneficiary]): Instance of CommercialBeneficiary representing the beneficiary information.
            dropped_group_categories (List[Type[Category]]): List of Category objects of the HCCs dropped by groups.

        Returns:
            List[Type[Category]]: List of Category objects representing the disease interactions.

        Notes:
            The way the SAS code works is applying the group categories before determine count categories. However, some of the
            interactions with RXC rely on HCCs that can get dropped as part of a group. Thus, the dropped_group_categories
            are passed in as well and considered.
        """
        category_list = [
            category.category
            for category in categories
            if category.type in ["disease", "group", "rx"]
        ]

        severe_illness, transplant = self._determine_severe_illness_transplant_status(
            category_list
        )
        # RXC categories are excluded
        category_count = len(
            [
                category
                for category in category_list
                if not category.startswith("RXC") and category != "HHS_HCC022"
            ]
        )

        if beneficiary.risk_model_age_group == "Child":
            interaction_dict = self._determine_child_interactions(
                severe_illness, transplant, category_count
            )
        else:
            # For adults, the category_list needs to be extended to include dropped
            # group categories
            category_list.extend(
                [category.category for category in dropped_group_categories]
            )

            interaction_dict = self._determine_adult_interactions(
                category_list, severe_illness, transplant, category_count, beneficiary
            )

        interactions = [
            Category(
                reference_files=self.model_group_reference_files,
                risk_model_population=beneficiary.risk_model_population,
                category=category,
                mapper_codes=trigger_codes,
            )
            for category, trigger_codes in interaction_dict.items()
        ]
        interactions.extend(categories)

        return interactions

    def _determine_child_interactions(
        self, severe_illness: bool, transplant: bool, category_count: int
    ):
        """
        Determines the interaction categories for a child based on severe illness and transplant status.

        This function evaluates whether a child has severe illness or has undergone a transplant,
        and assigns interaction categories accordingly. It creates a dictionary of interaction
        categories based on the severity and the number of diagnosis categories.

        Args:
            severe_illness (bool): Indicates whether the child has a severe illness.
            transplant (bool): Indicates whether the child has undergone a transplant.
            category_count (int): The number of diagnosis categories associated with the child.

        Returns:
            dict: A dictionary with the interaction categories as keys and None as values.
                Possible keys include severe illness categories based on category count and
                a transplant category if applicable.

        Notes:
            - Severe illness categories are defined based on the category count:
                - "SEVERE_HCC_COUNT{count}" for counts <= 5
                - "SEVERE_HCC_COUNT6_7" for counts between 6 and 7
                - "SEVERE_HCC_COUNT8PLUS" for counts >= 8
            - A transplant category "TRANSPLANT_HCC_COUNT4PLUS" is assigned if the category count is 4 or more.
        """
        # Children only get the severe illness interactions and transplant

        interaction_dict = {}
        severe_category = None
        transplant_category = None
        if severe_illness:
            if category_count <= 5:
                severe_category = f"SEVERE_HCC_COUNT{category_count}"
            elif 6 <= category_count <= 7:
                severe_category = "SEVERE_HCC_COUNT6_7"
            elif category_count >= 8:
                severe_category = "SEVERE_HCC_COUNT8PLUS"
            # In the case of severe and count not including any trigger codes
            interaction_dict[severe_category] = None

        if transplant:
            if category_count >= 4:
                transplant_category = "TRANSPLANT_HCC_COUNT4PLUS"
                interaction_dict[transplant_category] = None

        return interaction_dict

    def _determine_infant_maturity_status(self, category_list, beneficiary):
        """
        Determines the maturity status of an infant based on their age and associated diagnosis categories.

        This function assesses the maturity status of an infant beneficiary by considering their age and
        the presence of specific diagnosis categories. It assigns a maturity status that reflects the
        infant's developmental stage.

        Args:
            category_list (List[str]): A list of diagnosis categories associated with the beneficiary.
            beneficiary (CommercialBeneficiary): An object representing the beneficiary being scored, containing demographic
                                                and other relevant information, including age.

        Returns:
            str: The maturity status of the infant. Possible values include:
                - "Age1" for infants aged 1 year or if no relevant categories are present
                - "Extremely_Immature" for specific categories indicating extreme immaturity
                - "Immature" for specific categories indicating immaturity
                - "Premature_Multiples" for categories indicating premature multiples
                - "Term" for the category indicating a term birth

        Notes:
            - The maturity status is determined based on the presence of specific diagnosis categories
            in the category_list.
            - If the infant is 1 year old, the maturity status is set to "Age1" regardless of the categories.
            - If no relevant categories are present in the category_list, the maturity status defaults to "Age1".
              In this situation, the beneficiary risk_model_age is set to 1 for alignment as well.
        """
        maturity_status = None

        # If infant is 1, then maturity status is Age1
        # else it checks the categories associated with birth to determine status
        if beneficiary.age == 1:
            maturity_status = "Age1"
        else:
            if any(
                category in category_list
                for category in ["HHS_HCC242", "HHS_HCC243", "HHS_HCC244"]
            ):
                maturity_status = "Extremely_Immature"
            elif any(
                category in category_list for category in ["HHS_HCC245", "HHS_HCC246"]
            ):
                maturity_status = "Immature"
            elif any(
                category in category_list for category in ["HHS_HCC247", "HHS_HCC248"]
            ):
                maturity_status = "Premature_Multiples"
            elif "HHS_HCC249" in category_list:
                maturity_status = "Term"
            elif all(
                category not in category_list
                for category in [
                    "HHS_HCC242",
                    "HHS_HCC243",
                    "HHS_HCC244",
                    "HHS_HCC245",
                    "HHS_HCC246",
                    "HHS_HCC247",
                    "HHS_HCC248",
                    "HHS_HCC249",
                ]
            ):
                maturity_status = "Age1"
                beneficiary.risk_model_age = 1

        return maturity_status

    def _determine_infant_severity_level(self, category_list: List[str]):
        """
        Determines the severity level of an infant based on their associated diagnosis categories.

        This function evaluates the severity level of an infant beneficiary by checking the presence
        of specific diagnosis categories in the category_list. The severity level is assigned based
        on predefined severity categories.

        Args:
            category_list (List[str]): A list of diagnosis categories associated with the beneficiary.

        Returns:
            str: The severity level of the infant. Possible values include:
                - "Severity5" for the highest severity level
                - "Severity4" for high severity level
                - "Severity3" for moderate severity level
                - "Severity2" for low severity level
                - "Severity1" for the lowest severity level (default)

        Notes:
            - The function checks the presence of diagnosis categories in the following order:
            Severity5, Severity4, Severity3, Severity2, and Severity1.
            - The first matching severity level determines the severity status of the infant.
            - If no matching categories are found, the default severity status is "Severity1".
        """
        severity_5_hccs = [
            "HHS_HCC008",
            "HHS_HCC018",
            "HHS_HCC034",
            "HHS_HCC041",
            "HHS_HCC042",
            "HHS_HCC125",
            "HHS_HCC128",
            "HHS_HCC129",
            "HHS_HCC130",
            "HHS_HCC137",
            "HHS_HCC158",
            "HHS_HCC183",
            "HHS_HCC184",
            "HHS_HCC251",
        ]

        severity_4_hccs = [
            "HHS_HCC002",
            "HHS_HCC009",
            "HHS_HCC026",
            "HHS_HCC030",
            "HHS_HCC035_1",
            "HHS_HCC035_2",
            "HHS_HCC064",
            "HHS_HCC067",
            "HHS_HCC068",
            "HHS_HCC073",
            "HHS_HCC106",
            "HHS_HCC107",
            "HHS_HCC111",
            "HHS_HCC112",
            "HHS_HCC115",
            "HHS_HCC122",
            "HHS_HCC126",
            "HHS_HCC127",
            "HHS_HCC131",
            "HHS_HCC135",
            "HHS_HCC138",
            "HHS_HCC145",
            "HHS_HCC146",
            "HHS_HCC154",
            "HHS_HCC156",
            "HHS_HCC163",
            "HHS_HCC187",
            "HHS_HCC253",
        ]

        severity_3_hccs = [
            "HHS_HCC001",
            "HHS_HCC003",
            "HHS_HCC006",
            "HHS_HCC010",
            "HHS_HCC011",
            "HHS_HCC012",
            "HHS_HCC027",
            "HHS_HCC045",
            "HHS_HCC054",
            "HHS_HCC055",
            "HHS_HCC061",
            "HHS_HCC063",
            "HHS_HCC066",
            "HHS_HCC074",
            "HHS_HCC075",
            "HHS_HCC081",
            "HHS_HCC082",
            "HHS_HCC083",
            "HHS_HCC084",
            "HHS_HCC096",
            "HHS_HCC108",
            "HHS_HCC109",
            "HHS_HCC110",
            "HHS_HCC113",
            "HHS_HCC114",
            "HHS_HCC117",
            "HHS_HCC119",
            "HHS_HCC121",
            "HHS_HCC132",
            "HHS_HCC139",
            "HHS_HCC142",
            "HHS_HCC149",
            "HHS_HCC150",
            "HHS_HCC159",
            "HHS_HCC218",
            "HHS_HCC223",
            "HHS_HCC226",
            "HHS_HCC228",
        ]

        severity_2_hccs = [
            "HHS_HCC004",
            "HHS_HCC013",
            "HHS_HCC019",
            "HHS_HCC020",
            "HHS_HCC021",
            "HHS_HCC023",
            "HHS_HCC028",
            "HHS_HCC029",
            "HHS_HCC036",
            "HHS_HCC046",
            "HHS_HCC047",
            "HHS_HCC048",
            "HHS_HCC056",
            "HHS_HCC057",
            "HHS_HCC062",
            "HHS_HCC069",
            "HHS_HCC070",
            "HHS_HCC097",
            "HHS_HCC120",
            "HHS_HCC151",
            "HHS_HCC153",
            "HHS_HCC160",
            "HHS_HCC161_1",
            "HHS_HCC162",
            "HHS_HCC188",
            "HHS_HCC217",
            "HHS_HCC219",
        ]

        severity_1_hccs = [
            "HHS_HCC037_1",
            "HHS_HCC037_2",
            "HHS_HCC071",
            "HHS_HCC102",
            "HHS_HCC103",
            "HHS_HCC118",
            "HHS_HCC161_2",
            "HHS_HCC234",
            "HHS_HCC254",
        ]

        # In 2025 HHS_HCC071 is a higher severity
        if self.model_year == 2025:
            severity_1_hccs.remove("HHS_HCC071")
            severity_2_hccs.append("HHS_HCC071")

        severity_status = None

        if any(category in category_list for category in severity_5_hccs):
            severity_status = "Severity5"
        elif any(category in category_list for category in severity_4_hccs):
            severity_status = "Severity4"
        elif any(category in category_list for category in severity_3_hccs):
            severity_status = "Severity3"
        elif any(category in category_list for category in severity_2_hccs):
            severity_status = "Severity2"
        elif any(category in category_list for category in severity_1_hccs):
            severity_status = "Severity1"
        else:
            severity_status = "Severity1"

        return severity_status

    def _determine_adult_interactions(
        self,
        category_list: List[str],
        severe_illness: bool,
        transplant: bool,
        category_count: int,
        beneficiary: Type[CommercialBeneficiary],
    ) -> dict:
        """
        Determines the interaction categories for an adult beneficiary based on their diagnosis categories, severe illness, transplant status, and enrollment duration.

        This function evaluates various health conditions and their combinations to determine specific interaction categories. These interactions are based on the presence of certain diagnosis categories, severe illness, transplant status, and enrollment duration.

        Args:
            category_list (List[str]): A list of diagnosis categories associated with the beneficiary.
            severe_illness (bool): Indicates whether the beneficiary has a severe illness.
            transplant (bool): Indicates whether the beneficiary has undergone a transplant.
            category_count (int): The number of diagnosis categories associated with the beneficiary.
            beneficiary (CommercialBeneficiary): An object representing the beneficiary being scored, containing demographic and other relevant information.

        Returns:
            dict: A dictionary where keys are interaction categories and values are lists of triggering diagnosis categories.

        Notes:
            - Interaction categories are determined based on combinations of diagnosis categories related to liver, kidney, intestine, diabetes, autoimmune, and lung conditions.
            - Severe illness and transplant interaction categories are included based on the category count.
            - Enrollment duration interaction categories are also considered.
            - If certain interaction conditions are met, specific interaction categories are added to the dictionary with the relevant triggering categories.
        """
        liver = [
            category
            for category in [
                "HHS_HCC034",
                "HHS_HCC035_1",
                "HHS_HCC035_2",
                "HHS_HCC036",
                "HHS_HCC037_1",
            ]
            if category in category_list
        ]

        kidney = [
            category
            for category in ["HHS_HCC183", "HHS_HCC184", "HHS_HCC187", "HHS_HCC188"]
            if category in category_list
        ]

        intestine = [
            category
            for category in ["HHS_HCC041", "HHS_HCC048"]
            if category in category_list
        ]

        diabetes = [
            category
            for category in ["HHS_HCC018", "HHS_HCC019", "HHS_HCC020", "HHS_HCC021"]
            if category in category_list
        ]

        autoimmune = [
            category
            for category in ["HHS_HCC056", "HHS_HCC057"]
            if category in category_list
        ]

        lung = [
            category
            for category in ["HHS_HCC158", "HHS_HCC159"]
            if category in category_list
        ]

        # Keys will be interaction, values will be list of triggering categories
        interactions_dict = {}
        if all(category in category_list for category in ["RXC_01", "HHS_HCC001"]):
            interactions_dict["RXC_01_x_HCC001"] = ["RXC_01", "HHS_HCC001"]

        if all(["RXC_02" in category_list, liver]):
            interactions_dict["RXC_02_x_HCC037_1_036_035_2_035_1_034"] = [
                "RXC_02"
            ] + liver

        if all(category in category_list for category in ["RXC_03", "HHS_HCC142"]):
            interactions_dict["RXC_03_x_HCC142"] = ["RXC_03", "HHS_HCC142"]

        if all(["RXC_04" in category_list, kidney]):
            interactions_dict["RXC_04_x_HCC184_183_187_188"] = ["RXC_04"] + kidney

        if all(["RXC_05" in category_list, intestine]):
            interactions_dict["RXC_05_x_HCC048_041"] = ["RXC_05"] + intestine

        if all(["RXC_06" in category_list, diabetes]):
            interactions_dict["RXC_06_x_HCC018_019_020_021"] = ["RXC_06"] + diabetes

        if all(["RXC_07" in category_list, diabetes]):
            interactions_dict["RXC_07_x_HCC018_019_020_021"] = ["RXC_07"] + diabetes

        if all(category in category_list for category in ["RXC_08", "HHS_HCC118"]):
            interactions_dict["RXC_08_x_HCC118"] = ["RXC_08", "HHS_HCC118"]

        if all(["RXC_09" in category_list, autoimmune, intestine]):
            interactions_dict["RXC_09_x_HCC056_057_and_048_041"] = (
                ["RXC_09"] + autoimmune + intestine
            )

        if all(category in category_list for category in ["RXC_09", "HHS_HCC056"]):
            interactions_dict["RXC_09_x_HCC056"] = ["RXC_09", "HHS_HCC056"]

        if all(category in category_list for category in ["RXC_09", "HHS_HCC057"]):
            interactions_dict["RXC_09_x_HCC057"] = ["RXC_09", "HHS_HCC057"]

        if all(["RXC_09" in category_list, intestine]):
            interactions_dict["RXC_09_x_HCC048_041"] = ["RXC_09"] + intestine

        if all(["RXC_10" in category_list, lung]):
            interactions_dict["RXC_10_x_HCC159_158"] = ["RXC_10"] + lung

        if severe_illness:
            if category_count <= 9:
                severe_category = f"SEVERE_HCC_COUNT{category_count}"
                interactions_dict[severe_category] = None
            elif category_count >= 10:
                severe_category = "SEVERE_HCC_COUNT10PLUS"
                interactions_dict[severe_category] = None

        if transplant:
            if 4 <= category_count <= 7:
                transplant_category = f"TRANSPLANT_HCC_COUNT{category_count}"
                interactions_dict[transplant_category] = None
            elif category_count >= 8:
                transplant_category = "TRANSPLANT_HCC_COUNT8PLUS"
                interactions_dict[transplant_category] = None

        # Now do enrollment duration
        enrollment_duration = self._determine_enrollment_duration_category(
            category_count, beneficiary
        )
        if enrollment_duration:
            interactions_dict[enrollment_duration] = None

        return interactions_dict

    def _determine_severe_illness_transplant_status(self, category_list: List[str]):
        """
        Determines the severe illness and transplant status of a beneficiary based on their categories.

        This function evaluates whether the beneficiary has a severe illness or has undergone a transplant by checking the presence of specific diagnosis categories in the category_list.

        Args:
            category_list (List[str]): A list of diagnosis categories associated with the beneficiary.

        Returns:
            tuple: A tuple containing two boolean values:
                - severe_illness (bool): True if the beneficiary has a severe illness, otherwise False.
                - transplant (bool): True if the beneficiary has undergone a transplant, otherwise False.

        Notes:
            The DIY documentation uses a mixture of both groups and categories. The underlying categories of the groups are
            also included in the list as a group can be added or removed for a given model year but the underlying
            HCCs are still valid. For example, HCC 18 and HCC 183 were their own standalone categories in 2023, however in 2024
            they are now part of G24. So by including HCC 18, HCC 183, and G24 in the below lists, the code will work for both
            2023 and 2024.
        """
        severe_illness = any(
            category in category_list
            for category in [
                "HHS_HCC002",
                "HHS_HCC003",
                "HHS_HCC004",
                "HHS_HCC006",
                "HHS_HCC023",
                "HHS_HCC034",
                "HHS_HCC041",
                "HHS_HCC042",
                "HHS_HCC096",
                "HHS_HCC121",
                "HHS_HCC122",
                "HHS_HCC125",
                # G13
                "HHS_HCC126",
                "HHS_HCC127",
                "G13",
                # G14
                "HHS_HCC128",
                "HHS_HCC129",
                "G14",
                ##
                "HHS_HCC135",
                "HHS_HCC145",
                "HHS_HCC156",
                "HHS_HCC158",
                "HHS_HCC163",
                "HHS_HCC218",
                "HHS_HCC223",
                "HHS_HCC251",
                # G24
                "HHS_HCC018",
                "HHS_HCC183",
                "G24",
            ]
        )
        transplant = any(
            category in category_list
            for category in [
                "HHS_HCC034",
                "HHS_HCC041",
                # G14
                "HHS_HCC128",
                "HHS_HCC129",
                "G14",
                "HHS_HCC158",
                "HHS_HCC251",
                # G24
                "HHS_HCC018",
                "HHS_HCC183",
                "G24",
            ]
        )

        return severe_illness, transplant

    def _determine_enrollment_duration_category(self, category_count: int, beneficiary):
        """
        Determines the enrollment duration category for a beneficiary based on their category count and enrollment duration.

        This function evaluates the enrollment duration category for a beneficiary if the category count is greater than 0 and the enrollment duration (in months) is 6 or less.

        Args:
            category_count (int): The number of diagnosis categories associated with the beneficiary.
            beneficiary: An object representing the beneficiary being scored, containing demographic and other relevant information, including enrollment months.

        Returns:
            Union[str, None]: The enrollment duration category if applicable, otherwise None.

        Notes:
            - The function returns an enrollment duration category in the format "HCC_ED{months}" if the beneficiary has more than 0 categories and their enrollment duration is 6 months or less.
        """
        enrollment_duration_category = None
        if category_count > 0:
            if beneficiary.enrollment_months <= 6:
                enrollment_duration_category = f"HCC_ED{beneficiary.enrollment_months}"
        return enrollment_duration_category

    def _determine_age_gender_category(self, age: int, gender: str) -> str:
        """
        Determines the demographic category based on age, gender, and population.

        Args:
            age (int): Age of the individual.
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            population (str): Beneficiary model population used for scoring

        Returns:
            str: Demographic category based on age, gender, and population.
        """

        if age < 2:
            if gender == "F":
                demographic_category = None
            else:
                demographic_category = f"Age{age}_Male"
        else:
            demo_category_ranges = [
                "2_4",
                "5_9",
                "10_14",
                "15_20",
                "21_24",
                "25_29",
                "30_34",
                "35_39",
                "40_44",
                "45_49",
                "50_54",
                "55_59",
                "60_GT",
            ]

            demographic_category_range = determine_age_band(age, demo_category_ranges)
            demographic_category = f"{gender}AGE_LAST_{demographic_category_range}"

        return demographic_category



from typing import List


def determine_age_band(age: int, age_ranges: List[str]):
    """
    Determine the age band of a given age based on a list of age ranges.

    Args:
        age (int): The age to determine the age band for.
        age_ranges (list): A list of age ranges represented as strings in the format
                           "lower_upper" or "lower_GT" where "lower" is the lower
                           bound of the age range, "upper" is the upper bound (exclusive)
                           if provided, and "GT" signifies greater than.

    Returns:
        str: The age band that the given age falls into. Returns None if no matching
             age band is found in the provided age ranges.

    Example:
        >>> determine_age_band(25, ['18_24', '25_34', '35_GT'])
        '25_34'
    """
    range = None

    for age_range in age_ranges:
        age_band = age_range.split("_")
        lower, upper = 0, 999
        if len(age_band) == 1:
            lower = int(age_band[0])
            upper = lower + 1
        elif age_band[1] == "GT":
            lower = int(age_band[0])
        else:
            lower = int(age_band[0])
            upper = int(age_band[1]) + 1
        if lower <= age < upper:
            range = age_range
            break

    return range



from dataclasses import dataclass
from typing import List, Union, Dict


@dataclass
class BaseScoringResult:
    """
    Represents the scoring result for a specific individual in a population.

    Attributes:
        gender (str): The gender of the individual.
        age (int, optional): The age of the individual.
        dob (str, optional): The date of birth of the individual.
        diagnosis_codes (List[str], optional): A list of diagnosis codes associated with the individual.
        year (int, optional): The year passed into the model.
        risk_model_age (int): The age used in the risk model calculation.
        risk_model_population (str): The population group used in the risk model calculation.
        model_version (str): The version of the scoring model used.
        model_year (int): The year of the scoring model.
        score_raw (float): The raw score before any adjustments.
        disease_score_raw (float): The raw disease score before any adjustments.
        demographic_score_raw (float): The raw demographic score before any adjustments.
        score (float): The final score after adjustments.
        disease_score (float): The final disease score after adjustments.
        demographic_score (float): The final demographic score after adjustments.
        category_list (List[str]): A list of categories associated with the individual.
        category_details (Dict[str, str]): A dictionary containing details about each category.

    Notes:
        - The 'dob' attribute is expected to be in the format 'YYYY-MM-DD'.
        - The 'category_details' dictionary should contain category names as keys
          and corresponding details as values.
    """

    gender: str
    age: Union[int, None]
    dob: Union[str, None]
    diagnosis_codes: Union[List[str], None]
    year: Union[int, None]
    risk_model_age: int
    risk_model_population: str
    model_version: str
    model_year: int
    score_raw: float
    disease_score_raw: float
    demographic_score_raw: float
    score: float
    disease_score: float
    demographic_score: float
    category_list: List[str]
    category_details: Dict[str, str]


@dataclass
class MedicareScoringResult(BaseScoringResult):
    """
    Represents the scoring result for a Medicare beneficiary.

    Attributes:
        gender (str): The gender of the individual.
        orec (str): The OREC (original reason for entitlement category) of the individual.
        medicaid (bool): Indicates whether the individual has Medicaid coverage.
        age (int, optional): The age of the individual.
        dob (str, optional): The date of birth of the individual.
        diagnosis_codes (List[str], optional): A list of diagnosis codes associated with the individual.
        year (int, optional): The year passed into the model.
        risk_model_age (int): The age used in the risk model calculation.
        risk_model_population (str): The population group used in the risk model calculation.
        model_version (str): The version of the scoring model used.
        model_year (int): The year of the scoring model.
        score_raw (float): The raw score before any adjustments.
        disease_score_raw (float): The raw disease score before any adjustments.
        demographic_score_raw (float): The raw demographic score before any adjustments.
        score (float): The final score after adjustments.
        disease_score (float): The final disease score after adjustments.
        demographic_score (float): The final demographic score after adjustments.
        category_list (List[str]): A list of categories associated with the individual.
        category_details (Dict[str, str]): A dictionary containing details about each category.
        orec (str): The OREC (original reason for entitlement category) of the individual.
        medicaid (bool): Indicates whether the individual has Medicaid coverage.
        population (str): The population group to which the individual belongs.
        coding_intensity_adjuster (float): The coding intensity adjuster applied to the score.
        normalization_factor (float): The normalization factor applied to the score.

    Notes:
        - The 'dob' attribute is expected to be in the format 'YYYY-MM-DD'.
        - The 'category_details' dictionary should contain category names as keys
          and corresponding details as values.
    """

    orec: str
    medicaid: bool
    population: str
    coding_intensity_adjuster: float
    normalization_factor: float


@dataclass
class CommercialScoringResult(BaseScoringResult):
    """
    Represents the scoring result for a commercial insurance beneficiary.

    Attributes:
        gender (str): The gender of the individual.
        age (int, optional): The age of the individual.
        dob (str, optional): The date of birth of the individual.
        diagnosis_codes (List[str], optional): A list of diagnosis codes associated with the individual.
        year (int, optional): The year passed into the model.
        risk_model_age (int): The age used in the risk model calculation.
        risk_model_population (str): The population group used in the risk model calculation.
        model_version (str): The version of the scoring model used.
        model_year (int): The year of the scoring model.
        score_raw (float): The raw score before any adjustments.
        disease_score_raw (float): The raw disease score before any adjustments.
        demographic_score_raw (float): The raw demographic score before any adjustments.
        score (float): The final score after adjustments.
        disease_score (float): The final disease score after adjustments.
        demographic_score (float): The final demographic score after adjustments.
        category_list (List[str]): A list of categories associated with the individual.
        category_details (Dict[str, str]): A dictionary containing details about each category.
        metal_level (str): The metal level of the individual's insurance plan.
        csr_indicator (int): The cost-sharing reduction indicator.
        enrollment_days (int): The number of days the individual has been enrolled.
        last_enrollment_date (str, optional): The last enrollment date of the individual.
        enrollment_months (float): The number of months the individual has been enrolled.
        csr_adjuster (float): The cost-sharing reduction adjuster applied to the score.
        dropped_category_list (List[str], optional): A list of categories dropped from the scoring.
        dropped_category_details (Dict[str, str], optional): A dictionary containing details about each dropped category.

    Notes:
        - The 'dob' attribute is expected to be in the format 'YYYY-MM-DD'.
        - The 'category_details' and 'dropped_category_details' dictionaries should contain category names as keys
          and corresponding details as values.
    """

    metal_level: str
    csr_indicator: int
    enrollment_days: int
    last_enrollment_date: Union[str, None]
    enrollment_months: float
    csr_adjuster: float
    dropped_category_list: Union[List[str], None]
    dropped_category_details: Union[Dict[str, str], None]



import importlib.resources
import os
from pathlib import Path
from typing import Union
from .reference_files_loader import ReferenceFilesLoader


class BaseModel:
    """
    Represents a base model for healthcare Risk Adjustment models. This should not be
    called directly.

    Attributes:
        lob (str): Line of Business (LOB) for the model.
        version (str): Version of the model.
        year (int): Year for which the model is implemented (default is None).
        model_year (int): The actual year of the model.
        data_directory (Path): Path to the directory containing model data.
        reference_files (ReferenceFilesLoader): Loader for reference files.
    """

    def __init__(self, lob: str, version: str, year: Union[int, None] = None):
        """
        Initializes a BaseModel with the provided parameters.

        Args:
            lob (str): Line of Business (LOB) for the model.
            version (str): Version of the model.
            year (int, optional): Year for which the model is implemented (default is None).
        """
        self.lob = lob
        self.version = version
        self.year = year
        self.model_year = self._get_model_year()
        self.data_directory = self._get_data_directory()
        self.reference_files = ReferenceFilesLoader(self.data_directory, lob)

    def _get_model_year(self) -> int:
        """
        Determine the model year based on the provided year or the most recent available year.
        If the year passed in is invalid, it raises a value error.

        Returns:
            int: The model year.

        Raises:
            FileNotFoundError: If the specified version directory or reference data
                                directory does not exist.
            ValueError: If the year passed in is not valid for the Line of Business (LOB) and version,
                        or if no year is passed and there are no valid years available.
        """
        data_dir = importlib.resources.files(
            "risk_adjustment_model.reference_data"
        ).joinpath(f"{self.lob}")
        dirs = os.listdir(data_dir / self.version)
        years = [int(dir) for dir in dirs]

        if not self.year:
            max_year = max(years)
        elif self.year not in years:
            raise ValueError(
                f"Input year is not valid for LOB: {self.lob}, version: {self.version}. Valid years are {years}"
            )
        else:
            max_year = self.year

        return max_year

    def _get_data_directory(self) -> Path:
        """
        Get the directory path to the reference data for the Medicare model.

        Returns:
            Path: The directory path to the reference data.
        """
        data_dir = importlib.resources.files(
            "risk_adjustment_model.reference_data"
        ).joinpath(f"{self.lob}")
        data_directory = data_dir / self.version / str(self.model_year)

        return data_directory



from typing import List, Union, Type
from .utilities import determine_age_band
from .medicare_model import MedicareModel
from .category import Category
from .beneficiary import MedicareBeneficiary


class MedicareModelV24(MedicareModel):
    """
    This class represents the V24 Community Model for Medicare. It inherits from the MedicareModel class.

    Methods:
        __init__: Initializes the MedicareModelV24 instance.

        Overwrites:
            _get_normalization_factor: Retrieves the normalization factor based on the model year.
            _age_sex_edits: Applies age and sex edits to diagnosis codes.
            _determine_disease_interactions: Determines disease interactions based on Category objects and beneficiary information.

        Included for clarity:
            _determine_payment_count_category: Determines the payment count category based on the number of categories provided.
            _determine_age_gender_category: Determines the demographic category based on age, gender, and population.
            _determine_demographic_interactions: Determines demographic interactions based on gender, disability status, and Medicaid enrollment.

        New:
            _age_sex_edit_1: Applies age and sex edit 1 to a diagnosis code.
            _age_sex_edit_2: Applies age and sex edit 2 to a diagnosis code.
            _age_sex_edit_3: Applies age and sex edit 3 to a diagnosis code.
    """

    def __init__(self, year: Union[int, None] = None):
        super().__init__("v24", year)
        self.normalization_factor = self._get_normalization_factor(self.model_year)

    def _get_normalization_factor(self, year: int) -> float:
        """
        CMS updates normalization factor each year to apply to the risk scores. See:
        https://www.commonwealthfund.org/publications/explainer/2024/mar/how-government-updates-payment-rates-medicare-advantage-plans

        This dictionary is updated each year to include the normalization factor from the final announcement.

        Returns:
            float: The normalization factor.
        """
        norm_factor_dict = {
            2020: 1.069,
            2021: 1.097,
            2022: 1.118,
            2023: 1.127,
            2024: 1.146,
            2025: 1.153,
        }
        try:
            normalization_factor = norm_factor_dict[year]
        except KeyError:
            normalization_factor = 1

        return normalization_factor

    def _age_sex_edits(
        self, gender: str, age: int, diagnosis_code: str
    ) -> Union[List[str], None]:
        """
        Wrapper method to apply all model specific age and sex edits for a diagnosis code to
        category mapping. These are found in the model software file named something like
        "V24I0ED1".

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            age (int): Age of the individual.
            diagnosis_code (str): Diagnosis code to apply edits.

        Returns:
            Union[List[str], None]: List of categories after applying edits, or None if no edits applied.
        """
        new_category = self._age_sex_edit_1(gender, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_2(age, diagnosis_code)
        if new_category:
            return new_category
        new_category = self._age_sex_edit_3(age, diagnosis_code)
        if new_category:
            return new_category

    def _age_sex_edit_1(self, gender: str, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 1 to a diagnosis code.

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if gender == "F" and dx_code in ["D66", "D67"]:
            return ["HCC48"]

    def _age_sex_edit_2(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 2 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if age < 18 and dx_code in [
            "J410",
            "J411",
            "J418",
            "J42",
            "J430",
            "J431",
            "J432",
            "J438",
            "J439",
            "J440",
            "J441",
            "J449",
            "J982",
            "J983",
        ]:
            return ["HCC112"]

    def _age_sex_edit_3(self, age: int, dx_code: str) -> Union[List[str], None]:
        """
        Apply age and sex edit 3 to a diagnosis code.

        Args:
            age (int): Age of the individual.
            dx_code (str): Diagnosis code to apply the edit.

        Returns:
            Union[List[str], None]: List of categories after applying the edit, or None if the edit is not applicable.
        """
        if (age < 6 or age > 18) and dx_code == "F3481":
            return ["NA"]

    def _determine_disease_interactions(
        self, categories: List[Type[Category]], beneficiary: Type[MedicareBeneficiary]
    ) -> List[Type[Category]]:
        """
        Determines disease interactions based on provided Category objects and beneficiary information.

        Args:
            categories (List[Type[Category]]): List of Category objects representing disease categories.
            beneficiary (Type[MedicareBeneficiary]): Instance of MedicareBeneficiary representing the beneficiary information.

        Returns:
            List[Type[Category]]: List of Category objects representing the disease interactions.
        """
        category_list = [
            category.category for category in categories if category.type == "disease"
        ]
        cancer_list = ["HCC8", "HCC9", "HCC10", "HCC11", "HCC12"]
        diabetes_list = ["HCC17", "HCC18", "HCC19"]
        card_resp_fail_list = ["HCC82", "HCC83", "HCC84"]
        g_copd_cf_list = ["HCC110", "HCC111", "HCC112"]
        renal_v24_list = ["HCC134", "HCC135", "HCC136", "HCC137", "HCC138"]
        g_substance_use_disorder_v24_list = ["HCC54", "HCC55", "HCC56"]
        g_pyshiatric_v24_list = ["HCC57", "HCC58", "HCC59", "HCC60"]
        pressure_ulcer_list = ["HCC157", "HCC158", "HCC159"]

        cancer = any(category in category_list for category in cancer_list)
        diabetes = any(category in category_list for category in diabetes_list)
        card_resp_fail = any(
            category in category_list for category in card_resp_fail_list
        )
        chf = "HCC85" in category_list
        g_copd_cf = any(category in category_list for category in g_copd_cf_list)
        renal_v24 = any(category in category_list for category in renal_v24_list)
        sepsis = "HCC2" in category_list
        g_substance_use_disorder_v24 = any(
            category in category_list for category in g_substance_use_disorder_v24_list
        )
        g_pyshiatric_v24 = any(
            category in category_list for category in g_pyshiatric_v24_list
        )
        pressure_ulcer = any(
            category in category_list for category in pressure_ulcer_list
        )
        hcc47 = "HCC47" in category_list
        hcc96 = "HCC96" in category_list
        hcc188 = "HCC188" in category_list
        hcc114 = "HCC114" in category_list
        hcc57 = "HCC57" in category_list
        hcc79 = "HCC79" in category_list

        interactions_dict = {
            "HCC47_gCancer": all([cancer, hcc47]),
            "DIABETES_CHF": all([diabetes, chf]),
            "CHF_gCopdCF": all([chf, g_copd_cf]),
            "HCC85_gRenal_V24": all([chf, renal_v24]),
            "gCopdCF_CARD_RESP_FAIL": all([g_copd_cf, card_resp_fail]),
            "HCC85_HCC96": all([chf, hcc96]),
            "gSubstanceUseDisorder_gPsych": all(
                [g_pyshiatric_v24, g_substance_use_disorder_v24]
            ),
            "SEPSIS_PRESSURE_ULCER": all([sepsis, pressure_ulcer]),
            "SEPSIS_ARTIF_OPENINGS": all([sepsis, hcc188]),
            "ART_OPENINGS_PRESS_ULCER": all([hcc188, pressure_ulcer]),
            "gCopdCF_ASP_SPEC_B_PNEUM": all([g_copd_cf, hcc114]),
            "ASP_SPEC_B_PNEUM_PRES_ULC": all([hcc114, pressure_ulcer]),
            "SEPSIS_ASP_SPEC_BACT_PNEUM": all([sepsis, hcc114]),
            "SCHIZOPHRENIA_gCopdCF": all([hcc57, g_copd_cf]),
            "SCHIZOPHRENIA_CHF": all([hcc57, chf]),
            "SCHIZOPHRENIA_SEIZURES": all([hcc57, hcc79]),
            "DISABLED_HCC85": all([beneficiary.disabled, chf]),
            "DISABLED_PRESSURE_ULCER": all([beneficiary.disabled, pressure_ulcer]),
            "DISABLED_HCC161": all([beneficiary.disabled, "HCC161" in category_list]),
            "DISABLED_HCC39": all([beneficiary.disabled, "HCC39" in category_list]),
            "DISABLED_HCC77": all([beneficiary.disabled, "HCC77" in category_list]),
            "DISABLED_HCC6": all([beneficiary.disabled, "HCC6" in category_list]),
        }
        interaction_list = [key for key, value in interactions_dict.items() if value]

        category_count = self._determine_payment_count_category(category_list)
        if category_count:
            interaction_list.append(category_count)

        interactions = [
            Category(self.reference_files, beneficiary.risk_model_population, category)
            for category in interaction_list
        ]
        interactions.extend(categories)

        return interactions

    def _determine_payment_count_category(self, categories: list) -> str:
        """
        Determines the payment count category based on the number of categories provided.

        Args:
            categories (list): List of categories.

        Returns:
            str: Payment count category.
        """
        category_count = len(categories)
        category = None
        if category_count > 9:
            category = "D10P"
        elif category_count > 0:
            category = f"D{category_count}"

        return category

    def _determine_age_gender_category(
        self, age: int, gender: str, population: str
    ) -> str:
        """
        Determines the demographic category based on age, gender, and population.

        Args:
            age (int): Age of the individual.
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            population (str): Beneficiary model population used for scoring

        Returns:
            str: Demographic category based on age, gender, and population.
        """
        if population[:2] == "NE":
            demo_category_ranges = [
                "0_34",
                "35_44",
                "45_54",
                "55_59",
                "60_64",
                "65",
                "66",
                "67",
                "68",
                "69",
                "70_74",
                "75_79",
                "80_84",
                "85_89",
                "90_94",
                "95_GT",
            ]
        else:
            demo_category_ranges = [
                "0_34",
                "35_44",
                "45_54",
                "55_59",
                "60_64",
                "65_69",
                "70_74",
                "75_79",
                "80_84",
                "85_89",
                "90_94",
                "95_GT",
            ]

        demographic_category_range = determine_age_band(age, demo_category_ranges)

        if population[:2] == "NE":
            demographic_category = f"NE{gender}{demographic_category_range}"
        else:
            demographic_category = f"{gender}{demographic_category_range}"

        return demographic_category

    def _determine_demographic_interactions(
        self, gender: str, orig_disabled: bool, medicaid: bool
    ) -> List[str]:
        """
        Determines demographic interactions based on gender, disability status, and Medicaid enrollment.

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            orig_disabled (bool): Indicates if the individual was originally disabled.
            medicaid (bool): Indicates if the individual is enrolled in Medicaid.

        Returns:
            List[str]: List of demographic interaction labels.
        """
        demo_interactions = []
        if gender == "F" and orig_disabled:
            demo_interactions.append("OriginallyDisabled_Female")
        elif gender == "M" and orig_disabled:
            demo_interactions.append("OriginallyDisabled_Male")

        if medicaid:
            demo_interactions.append("LTIMCAID")

        return demo_interactions



import copy
from typing import Union, Type, List, Tuple
from .utilities import determine_age_band
from .beneficiary import CommercialBeneficiary
from .category import Category
from .result import CommercialScoringResult
from .mapper import DxCodeCategory, NDCCodeCategory, ProcCodeCategory
from .model import BaseModel


class CommercialModel(BaseModel):
    """
    This is the foundation class for Commercial Models. It is not to be called directly. It loads all
    relevant information for that model and year as class attributes.

    How this class works:
    1.  Instantiate the class with set up information: model version and year (optional).
        If year is None, the code will determine the most recent year and use that.
    2.  Instantiating the class loads all reference information into memory for performance purposes.
    3.  The entry point is the "score" method.
    4.  Since each model has its own nuances, each model is its own class inheriting from this class.

    Attributes:
    reference_files_version_dict: Contains the version of reference files and

    Notes:
        In terms of inheritance for child classes and how it relates to methods, the code is
        organizaed in the following fashion:

        Methods strongly advised against overwriting:
            score
            _build_category_details
        Methods unlikely needing overwriting but could happen based on needs:
            _apply_hierarchies
            _apply_groups
            _determine_demographic_categories
            _get_dx_categories
            _get_ndc_categories
            _get_proc_categories
            _determine_infant_categories
        Methods likely needing overwriting, e.g.:
            _determine_infant_maturity_status
            _determine_infant_severity_level
            _determine_age_gender_category
            _determine_disease_interactions
            _age_sex_edits
        Helper methods to not override, e.g.: _apply_csr_adj

    For reference files, CMS historically releases 3 versions of the DIY software: July/August of Benefit Year (BY),
    December of Benefit Year or January of the following year (e.g. Dec 2024 or Jan 2025), then a final version at the
    close of the benefit year April BY+1 (so for BY2024 April 2025). Prior to the BY, CMS publishes the coefficient
    factors and any category/group changes. Given this cadence, the convention will be
    0.X = Anything before first DIY release
    1.X = First DIY release
    2.X = Second DIY release
    3.X = Third DIY release

    After the decmial would be for errata that CMS addresses or for bug fixes.
    """

    reference_files_version_dict = {
        9999: {
            "version": "placeholder",  # should be 0.0, 1.0, 2.0, 3.0, etc. with decimals only being used if there is an errata or fix to one of the versions
            "description": "placeholder",
        }
    }

    def __init__(self, version: str, year: Union[int, None] = None):
        super().__init__(lob="commercial", version=version, year=year)
        self.reference_files_version = self._get_reference_files_version(year)
        self.reference_files_description = self._get_reference_description(year)
        # The way the Commercial Model is designed is that the model determines categories
        # and score is based on the age group in addition to the diagnosis codes.
        # To handle this the category map contains as a key agegroup_hcc, this map now
        # needs to be pared down to the relevant model now that we know the age group
        # associated with the scoring run
        self.model_group_reference_files = copy.deepcopy(self.reference_files)

    def _get_reference_files_version(self, year: Union[int, None]) -> str:
        """
        Get the reference version for the specified year.

        Args:
            year (int, optional): The year for which to get the reference version.

        Returns:
            str: The reference version for the specified year.
        """
        return self.reference_files_version_dict.get(year, {}).get("version", "Unknown")

    def _get_reference_description(self, year: Union[int, None]) -> str:
        """
        Get the reference description for the specified year.

        Args:
            year (int, optional): The year for which to get the reference description.

        Returns:
            str: The reference description for the specified year.
        """
        return self.reference_files_version_dict.get(year, {}).get(
            "description", "No description available"
        )

    def score(
        self,
        gender: str,
        metal_level: str,
        csr_indicator: int,
        enrollment_days: int,
        diagnosis_codes: Union[List[str], None] = None,
        ndc_codes: Union[List[str], None] = None,
        proc_codes: Union[List[str], None] = None,
        age: Union[int, None] = None,
        dob: Union[str, None] = None,
        last_enrollment_date: Union[str, None] = None,
        verbose: bool = False,
    ) -> Type[CommercialScoringResult]:
        """
        Determines the risk score for the inputs. Entry point for end users.

        Steps:
        1. Use beneficiary information to get the demographic categories
        2. Using diagnosis code, ndc, and procedure code inputs and beneficiary information
           to get the code to category relationship
        3. Get the unique set of categories from the codes
        4. Apply hierarchies
        5. Determine disease interactions
        6. Apply groups

        Args:
            gender (str): Gender of the beneficiary being scored, valid values M or F.
            metal_level (str): Metal level of the beneficiary's insurance plan.
            csr_indicator (int): Cost-sharing reduction indicator.
            enrollment_days (int): Number of days the beneficiary has been enrolled.
            diagnosis_codes (list, optional): List of the diagnosis codes associated with the beneficiary, can be None.
            ndc_codes (list, optional): List of National Drug Codes (NDC) associated with the beneficiary, can be None.
            proc_codes (list, optional): List of procedure codes associated with the beneficiary, can be None.
            age (int, optional): Age of the beneficiary, can be None.
            dob (str, optional): Date of birth of the beneficiary, can be None.
            last_enrollment_date (str, optional): Last enrollment date of the beneficiary, can be None.
            verbose (bool): Indicates if trimmed output or full output is desired.

        Returns:
            ScoringResult: An instantiated object of ScoringResult class.
        """
        beneficiary = CommercialBeneficiary(
            gender,
            metal_level,
            enrollment_days,
            csr_indicator,
            age,
            dob,
            self.model_year,
            last_enrollment_date,
        )
        self.csr_adjuster = self._get_csr_adjuster(self.model_year, csr_indicator)

        # The way the Commercial Model is designed is that the model determines categories
        # and score is based on the age group in addition to the diagnosis codes.
        # To handle this the category map contains as a key agegroup_hcc, this map now
        # needs to be pared down to the relevant model now that we know the age group
        # associated with the scoring run
        self.model_group_reference_files.category_weights = {
            key.split("_", 1)[1]: self.reference_files.category_weights[key]
            for key in self.reference_files.category_weights.keys()
            if key.startswith(beneficiary.risk_model_age_group)
        }
        self.model_group_reference_files.group_definitions = (
            self.reference_files.group_definitions.get(beneficiary.risk_model_age_group)
        )

        demo_categories = self._determine_demographic_categories(beneficiary)

        if diagnosis_codes or ndc_codes or proc_codes:
            mapper_categories = []
            cat_dict = {}
            if diagnosis_codes:
                dx_categories = self._get_dx_categories(diagnosis_codes, beneficiary)
                mapper_categories.extend(dx_categories)
            # Only Adults get RXCs
            if ndc_codes and beneficiary.risk_model_age_group == "Adult":
                ndc_categories = self._get_ndc_categories(ndc_codes)
                mapper_categories.extend(ndc_categories)
            if proc_codes and beneficiary.risk_model_age_group == "Adult":
                proc_categories = self._get_proc_categories(proc_codes)
                mapper_categories.extend(proc_categories)

            # Some diagnosis codes go to more than one category thus the category is a list
            # and it is a two step process to unpack them
            unique_disease_cats = set(
                category
                for mapper_code in mapper_categories
                for category in mapper_code.categories
                if category is not None and category != "NA"
            )

            for category in unique_disease_cats:
                # This is done to obtain a consistent output for diagnosis map
                # If there are no diagnosis codes mapping to the category it should
                # return None, as opposed to an empty list. The below statement making
                # the map would create an empty list, thus the if statement following
                # to catch that and modify it
                mapper_map = [
                    mapper_code.mapper_code
                    for mapper_code in mapper_categories
                    if category in mapper_code.categories
                ]
                cat_dict[category] = mapper_map
        else:
            cat_dict = {}
            unique_disease_cats = None

        if unique_disease_cats:
            unique_categories = demo_categories + list(unique_disease_cats)
        else:
            unique_categories = demo_categories

        # Create a placeholder for this to be used later
        dropped_categories = []

        # At this point the model groups start diverging thus a separate method for
        # infants
        if beneficiary.risk_model_age_group == "Infant":
            # Since the infant model does not have HHS_HCCs with weights, not instantiating
            # categories yet
            categories = self._determine_infant_categories(
                unique_categories, beneficiary
            )
        else:
            # Adult and Child both apply hierarchies and groups and have
            # interactions. Children do not have RX categories, and interactions
            # are different per group and will need to be handled in other methods
            if beneficiary.risk_model_age_group == "Child":
                categories = [
                    Category(
                        self.model_group_reference_files,
                        beneficiary.risk_model_population,
                        category,
                        cat_dict.get(category),
                    )
                    for category in unique_categories
                    if "RX" not in category
                ]
            else:
                categories = [
                    Category(
                        self.model_group_reference_files,
                        beneficiary.risk_model_population,
                        category,
                        cat_dict.get(category),
                    )
                    for category in unique_categories
                ]
            categories, dropped_hierarchy_categories = self._apply_hierarchies(
                categories
            )
            categories, dropped_group_categories = self._apply_groups(
                categories, beneficiary
            )
            # Count is based on categories and groups, so interactions occur after
            # groups, however some interactions are based on categories dropped by
            # groups, so the dropped categories are also passed in
            categories = self._determine_interactions(
                categories, beneficiary, dropped_group_categories
            )

            # Add the dropped categories to their own list to return in results
            if dropped_hierarchy_categories:
                dropped_categories.extend(dropped_hierarchy_categories)
            if dropped_group_categories:
                dropped_categories.extend(dropped_group_categories)

        score_raw = sum([category.coefficient for category in categories])
        disease_score_raw = sum(
            [
                category.coefficient
                for category in categories
                if "disease" in category.type
            ]
        )
        demographic_score_raw = sum(
            [
                category.coefficient
                for category in categories
                if "demographic" in category.type
            ]
        )

        category_details = self._build_category_details(categories, verbose)

        if dropped_categories:
            dropped_category_details = self._build_category_details(
                dropped_categories, verbose
            )
        else:
            dropped_category_details = None

        results = CommercialScoringResult(
            gender=beneficiary.gender,
            metal_level=beneficiary.metal_level,
            enrollment_days=beneficiary.enrollment_days,
            csr_indicator=beneficiary.csr_indicator,
            age=beneficiary.age,
            dob=beneficiary.dob,
            last_enrollment_date=beneficiary.last_enrollment_date,
            enrollment_months=beneficiary.enrollment_months,
            diagnosis_codes=diagnosis_codes,
            year=self.year,
            risk_model_age=beneficiary.risk_model_age,
            risk_model_population=beneficiary.risk_model_population,
            model_version=self.version,
            model_year=self.model_year,
            csr_adjuster=self.csr_adjuster,
            score_raw=score_raw,
            disease_score_raw=disease_score_raw,
            demographic_score_raw=demographic_score_raw,
            score=self._apply_csr_adj(score_raw),
            disease_score=self._apply_csr_adj(disease_score_raw),
            demographic_score=self._apply_csr_adj(demographic_score_raw),
            category_list=[category.category for category in categories],
            category_details=category_details,
            dropped_category_list=[
                category.category for category in dropped_categories
            ],
            dropped_category_details=dropped_category_details,
        )

        return results

    def _build_category_details(
        self, categories: List[Type[Category]], verbose: bool
    ) -> dict:
        """
        Constructs a dictionary with details about each category based on the given list of Category objects.

        Args:
            categories (list[Category]): A list of Category objects to build the details from.
            verbose (bool): Flag indicating whether to include detailed information in the output.

        Returns:
            dict: A dictionary with category details. If verbose is True, includes extended information
                such as type, category number, description, and dropped categories. Otherwise, includes
                only coefficient and trigger code map.

        The structure of the returned dictionary is as follows:
            {
                "category_name": {
                    "coefficient": float,
                    "type": str,  # Included only if verbose is True
                    "category_number": int,  # Included only if verbose is True
                    "category_description": str,  # Included only if verbose is True
                    "dropped_categories": list,  # Included only if verbose is True
                    "trigger_code_map": dict,
                },
                ...
            }
        """
        # Combine the dictionaries to make output
        category_details = {}
        for category in categories:
            if verbose:
                category_details[category.category] = {
                    "coefficient": category.coefficient,
                    "type": category.type,
                    "category_number": category.number,
                    "category_description": category.description,
                    "dropped_categories": category.dropped_categories,
                    "trigger_code_map": category.mapper_codes,
                }
            else:
                category_details[category.category] = {
                    "coefficient": category.coefficient,
                    "trigger_code_map": category.mapper_codes,
                }

        return category_details

    # --- Methods which may need to be overwritten but unlikely to be overwritten ---

    def _determine_demographic_categories(
        self, beneficiary: Type[CommercialBeneficiary]
    ) -> List[str]:
        """
        Determine demographic categories based on beneficiary attributes.

        Args:
            beneficiary (Type[CommercialBeneficiary]): An instance of CommercialBeneficiary.

        Returns:
            list: A list containing demographic categories.
        """
        demo_cats = []
        demo_cat = self._determine_age_gender_category(
            beneficiary.age, beneficiary.gender
        )
        # Infant females do not have a weighted demographic category thus this check is needed
        if demo_cat:
            demo_cats.append(demo_cat)
        return demo_cats

    def _determine_infant_categories(
        self, unique_categories: List[str], beneficiary: Type[CommercialBeneficiary]
    ) -> List[Type[Category]]:
        """
        Determines the categories for an infant beneficiary, combining demographic and interaction categories
        based on the severity level and maturity status derived from their diagnosis codes.

        The function integrates multiple steps specific to the Infant model, which differs from other models
        by combining demographic components with interaction categories. These interaction categories are determined
        by the severity level and maturity status derived from the diagnosis codes.

        Args:
            unique_categories (List[str]): A list of unique categories derived from the beneficiary's diagnosis codes.
            beneficiary (CommercialBeneficiary): An object representing the beneficiary being scored, containing demographic
                                                and other relevant information.

        Returns:
            List[Category]: A list of Category objects representing the combined demographic and interaction categories for the infant.

        The function follows these steps:
        1. Creates a list of Category objects based on the demographic categories from the unique_categories list, excluding those starting with "HHS_HCC".
        2. Determines the maturity status and severity level of the infant based on the unique categories.
        3. If both maturity status and severity level are determined, creates and adds an interaction category combining both.
        4. Handles a specific scenario where an infant's age is 0, but the maturity level indicates Age1, ensuring the correct demographic category is assigned.
        """
        # If infants don't have disease categories, can't compute the maturity and
        # severity level, so first build the categories list based on demographic
        # then consider the ones based on categories
        categories = [
            Category(
                self.model_group_reference_files,
                beneficiary.risk_model_population,
                category,
            )
            for category in unique_categories
            if not category.startswith("HHS_HCC")
        ]

        maturity_status = self._determine_infant_maturity_status(
            unique_categories, beneficiary
        )
        severity_level = self._determine_infant_severity_level(unique_categories)
        if maturity_status and severity_level:
            interaction_category = f"{maturity_status}_x_{severity_level}"

            categories.append(
                Category(
                    self.model_group_reference_files,
                    beneficiary.risk_model_population,
                    interaction_category,
                )
            )

        # There is a scenario in the model where the age of an infant is 0, but the infant
        # has a maturity level of Age1. In this case, the infant is supposed to get the
        # Age1_Male demographic category.

        if (
            beneficiary.age == 0
            and beneficiary.gender == "M"
            and maturity_status == "Age1"
        ):
            # Append Age1_Male to category list
            categories.append(
                Category(
                    self.model_group_reference_files,
                    beneficiary.risk_model_population,
                    "Age1_Male",
                )
            )
            # Remove Age0_Male from list
            categories = [
                category
                for category in categories
                if category.category not in "Age0_Male"
            ]

        return categories

    def _apply_hierarchies(
        self, categories: List[Type[Category]]
    ) -> Tuple[List[Type[Category]], List[Type[Category]]]:
        """
        Filters out categories falling into hierarchies per the model hierarchy_definition file.

        Args:
            categories (List[Type[Category]]): List of category objects to process.

        Returns:
            List[Type[Category]]: List of category objects after filtering.

        Notes:
            For each category, the codes dropped are tracked and assigned to the
            attribute "dropped_categories" of the category object.
        """
        category_list = [category.category for category in categories]
        dropped_codes_total = []

        for category in categories:
            dropped_codes = []
            if (
                category.category
                in self.model_group_reference_files.hierarchy_definitions.keys()
            ):
                for (
                    remove_category
                ) in self.model_group_reference_files.hierarchy_definitions[
                    category.category
                ]["remove_code"]:
                    if remove_category in category_list:
                        dropped_codes.append(remove_category)
                        dropped_codes_total.append(remove_category)

            if dropped_codes:
                category.dropped_categories = dropped_codes

        final_categories = []
        dropped_categories = []

        # Remove the dropped codes and store them in a separate list
        for category in categories:
            if category.category not in dropped_codes_total:
                final_categories.append(category)
            else:
                dropped_categories.append(category)

        return final_categories, dropped_categories

    def _apply_groups(
        self, categories: List[Type[Category]], beneficiary: Type[CommercialBeneficiary]
    ) -> Tuple[List[Type[Category]], List[Type[Category]]]:
        """
        Assigns groups based on the "remove_codes" the member has and
        adds the group category and filters out the normal categories per the model
        group_definition file.

        Args:
            categories (List[Type[Category]]): List of category objects to process.

        Returns:
            List[Type[Category]]: List of category objects after filtering.

        Notes:
            For each category, the codes dropped are tracked and assigned to the
            attribute "dropped_categories" of the category object.

            It is possible for a group to drop two categories that are not in a
            hierarchy. As such it needs to be treated similarly to how applying
            hierarchies work.
        """
        category_list = [category.category for category in categories]
        dropped_codes_total = []
        group_dict = {}

        for category in category_list:
            if category in self.model_group_reference_files.group_definitions.keys():
                # At this point need to add the Group code and the code it drops
                # to a dictionary, and then check if that group code exists
                # if it does, then update the value, else add a new value
                dropped_codes_total.append(category)
                group_category = self.model_group_reference_files.group_definitions[
                    category
                ]

                if group_category in group_dict:
                    group_dict[group_category].append(category)
                else:
                    group_dict[group_category] = [category]

        final_categories = []
        dropped_categories = []

        # Remove the dropped codes and store them in a separate list
        for category in categories:
            if category.category not in dropped_codes_total:
                final_categories.append(category)
            else:
                dropped_categories.append(category)

        # Add groups to the list
        # Since "categories" trigger the code, they are both
        # the dropped codes and the mapper codes
        group_categories = [
            Category(
                reference_files=self.model_group_reference_files,
                risk_model_population=beneficiary.risk_model_population,
                category=category,
                mapper_codes=dropped_codes,
                dropped_categories=dropped_codes,
            )
            for category, dropped_codes in group_dict.items()
        ]

        if group_categories:
            final_categories.extend(group_categories)

        return final_categories, dropped_categories

    def _get_dx_categories(
        self, diagnosis_codes: List[str], beneficiary: Type[CommercialBeneficiary]
    ) -> List[Type[DxCodeCategory]]:
        """
        Generates diagnosis code to categories relationships based on provided diagnosis codes
        and beneficiary information.

        Args:
            diagnosis_codes (List[str]): List of diagnosis codes.
            beneficiary (Type[CommercialBeneficiary]): Instance of CommercialBeneficiary representing the beneficiary information.

        Returns:
            List[Type[DxCodeCategory]]: List of DxCodeCategory objects representing the diagnosis code categories.
        """
        dx_categories = [
            DxCodeCategory(
                self.model_group_reference_files.category_map, diagnosis_code
            )
            for diagnosis_code in diagnosis_codes
        ]

        for dx in dx_categories:
            edit_category = self._age_sex_edits(
                beneficiary.gender, beneficiary.age, dx.mapper_code
            )
            if edit_category:
                dx.categories = edit_category

        return dx_categories

    def _get_ndc_categories(self, ndc_codes: List[str]) -> List[Type[NDCCodeCategory]]:
        """
        Generates diagnosis code to categories relationships based on provided diagnosis codes
        and beneficiary information.

        Args:
            diagnosis_codes (List[str]): List of diagnosis codes.
            beneficiary (Type[CommercialBeneficiary]): Instance of CommercialBeneficiary representing the beneficiary information.

        Returns:
            List[Type[DxCodeCategory]]: List of DxCodeCategory objects representing the diagnosis code categories.
        """
        ndc_categories = [
            NDCCodeCategory(self.model_group_reference_files.category_map, ndc_code)
            for ndc_code in ndc_codes
        ]

        return ndc_categories

    def _get_proc_categories(
        self, proc_codes: List[str]
    ) -> List[Type[ProcCodeCategory]]:
        """
        Generates diagnosis code to categories relationships based on provided diagnosis codes
        and beneficiary information.

        Args:
            diagnosis_codes (List[str]): List of diagnosis codes.
            beneficiary (Type[CommercialBeneficiary]): Instance of CommercialBeneficiary representing the beneficiary information.

        Returns:
            List[Type[DxCodeCategory]]: List of DxCodeCategory objects representing the diagnosis code categories.
        """
        proc_code_categories = [
            ProcCodeCategory(self.model_group_reference_files.category_map, proc_code)
            for proc_code in proc_codes
        ]

        return proc_code_categories

    def _get_csr_adjuster(self, year: int, csr_indicator: str) -> float:
        """
        Gets the appropriate CSR adjuster for the model year as outlined by
        CMS in their DIY Technical PDF Guidance:
        https://www.cms.gov/marketplace/resources/regulations-guidance

        Returns:
            float: The coding intensity adjuster.
        """
        csr_adjuster_dict = {
            2019: {"1": 1.00, "2": 1.07, "3": 1.12, "4": 1.15},
            2020: {"1": 1.00, "2": 1.07, "3": 1.12, "4": 1.15},
            2021: {"1": 1.00, "2": 1.07, "3": 1.12, "4": 1.15},
            2022: {"1": 1.00, "2": 1.07, "3": 1.12, "4": 1.15},
            2023: {"1": 1.00, "2": 1.07, "3": 1.12, "4": 1.15},
            2024: {"1": 1.00, "2": 1.07, "3": 1.12, "4": 1.15},
            2025: {"1": 1.00, "2": 1.07, "3": 1.12, "4": 1.15},
        }

        if csr_adjuster_dict.get(year):
            csr_adjuster = csr_adjuster_dict[year].get(csr_indicator, 1.00)
        else:
            csr_adjuster = 1.00

        return csr_adjuster

    # --- Methods likely to be overwritten by each model class ---

    def _determine_age_gender_category(self, age: int, gender: str) -> str:
        """
        Determines the demographic category based on age, gender, and population.

        Args:
            age (int): Age of the individual.
            gender (str): Gender of the individual ('M' for male, 'F' for female).

        Returns:
            str: Demographic category based on age and gender.
        """

        if age < 2:
            if gender == "F":
                demographic_category = None
            else:
                demographic_category = f"Age{age}_Male"
        else:
            demo_category_ranges = [
                "2_4",
                "5_9",
                "10_14",
                "15_20",
                "21_24",
                "25_29",
                "30_34",
                "35_39",
                "40_44",
                "45_49",
                "50_54",
                "55_59",
                "60_GT",
            ]

            demographic_category_range = determine_age_band(age, demo_category_ranges)
            demographic_category = f"{gender}AGE_LAST_{demographic_category_range}"

        return demographic_category

    def _determine_interactions(
        self,
        categories: List[Type[Category]],
        beneficiary: Type[CommercialBeneficiary],
        dropped_group_categories: List[Type[Category]],
    ) -> List[Type[Category]]:
        """
        Determines disease interactions based on provided Category objects and beneficiary information.
        Placeholder to be overwritten by child classes.

        Args:
            categories (List[Type[Category]]): List of Category objects representing disease categories.
            beneficiary (Type[CommercialBeneficiary]): Instance of CommercialBeneficiary representing the beneficiary information.

        Returns:
            List[Type[Category]]: List of Category objects representing the disease interactions.
        """
        interaction_list = []

        interactions = [
            Category(
                self.model_group_reference_files,
                beneficiary.risk_model_population,
                beneficiary.risk_model_age_group,
                category,
            )
            for category in interaction_list
        ]

        interactions.append(categories)

        return interactions

    def _age_sex_edits(
        self, gender: str, age: int, diagnosis_code: str
    ) -> Union[List[str], None]:
        """
        Placeholder function to be overwritten by child clasess. This to encapsulate
        the age sex edits for a model that are to be performed on the
        DxCodeCategory objects in the _get_dx_categories method.

        Returns:
            List[str]: List of categories based on input gender, age, diagnosis code
        """
        return ["NA"]

    def _determine_infant_maturity_status(
        self, category_list: List[str], beneficiary: Type[CommercialBeneficiary]
    ):
        """
        Determines the maturity status of an infant based on their age and associated diagnosis categories.

        This function assesses the maturity status of an infant beneficiary by considering their age and
        the presence of specific diagnosis categories. It assigns a maturity status that reflects the
        infant's developmental stage.

        Args:
            category_list (List[str]): A list of diagnosis categories associated with the beneficiary.
            beneficiary (CommercialBeneficiary): An object representing the beneficiary being scored, containing demographic
                                                and other relevant information, including age.

        Returns:
            str: The maturity status of the infant. Possible values include:
                - "Age1" for infants aged 1 year or if no relevant categories are present
                - "Extremely_Immature" for specific categories indicating extreme immaturity
                - "Immature" for specific categories indicating immaturity
                - "Premature_Multiples" for categories indicating premature multiples
                - "Term" for the category indicating a term birth

        Notes:
            - The maturity status is determined based on the presence of specific diagnosis categories
            in the category_list.
            - If the infant is 1 year old, the maturity status is set to "Age1" regardless of the categories.
            - If no relevant categories are present in the category_list, the maturity status defaults to "Age1".
        """
        maturity_status = None

        # If infant is 1, then maturity status is Age1
        # else it checks the categories associated with birth to determine status
        if beneficiary.age == 1:
            maturity_status = "Age1"
        else:
            if any(
                category in category_list
                for category in ["HHS_HCC242", "HHS_HCC243", "HHS_HCC244"]
            ):
                maturity_status = "Extremely_Immature"
            elif any(
                category in category_list for category in ["HHS_HCC245", "HHS_HCC246"]
            ):
                maturity_status = "Immature"
            elif any(
                category in category_list for category in ["HHS_HCC247", "HHS_HCC248"]
            ):
                maturity_status = "Premature_Multiples"
            elif "HHS_HCC249" in category_list:
                maturity_status = "Term"
            elif all(
                category not in category_list
                for category in [
                    "HHS_HCC242",
                    "HHS_HCC243",
                    "HHS_HCC244",
                    "HHS_HCC245",
                    "HHS_HCC246",
                    "HHS_HCC247",
                    "HHS_HCC248",
                    "HHS_HCC249",
                ]
            ):
                maturity_status = "Age1"

        return maturity_status

    def _determine_infant_severity_level(self, category_list: List[str]):
        """
        Determines the severity level of an infant based on their associated diagnosis categories.

        This function evaluates the severity level of an infant beneficiary by checking the presence
        of specific diagnosis categories in the category_list. The severity level is assigned based
        on predefined severity categories.

        Args:
            category_list (List[str]): A list of diagnosis categories associated with the beneficiary.

        Returns:
            str: The severity level of the infant. Possible values include:
                - "Severity5" for the highest severity level
                - "Severity4" for high severity level
                - "Severity3" for moderate severity level
                - "Severity2" for low severity level
                - "Severity1" for the lowest severity level (default)

        Notes:
            - The function checks the presence of diagnosis categories in the following order:
            Severity5, Severity4, Severity3, Severity2, and Severity1.
            - The first matching severity level determines the severity status of the infant.
            - If no matching categories are found, the default severity status is "Severity1".
        """
        severity_5_hccs = [
            "HHS_HCC008",
            "HHS_HCC018",
            "HHS_HCC034",
            "HHS_HCC041",
            "HHS_HCC042",
            "HHS_HCC125",
            "HHS_HCC128",
            "HHS_HCC129",
            "HHS_HCC130",
            "HHS_HCC137",
            "HHS_HCC158",
            "HHS_HCC183",
            "HHS_HCC184",
            "HHS_HCC251",
        ]

        severity_4_hccs = [
            "HHS_HCC002",
            "HHS_HCC009",
            "HHS_HCC026",
            "HHS_HCC030",
            "HHS_HCC035_1",
            "HHS_HCC035_2",
            "HHS_HCC064",
            "HHS_HCC067",
            "HHS_HCC068",
            "HHS_HCC073",
            "HHS_HCC106",
            "HHS_HCC107",
            "HHS_HCC111",
            "HHS_HCC112",
            "HHS_HCC115",
            "HHS_HCC122",
            "HHS_HCC126",
            "HHS_HCC127",
            "HHS_HCC131",
            "HHS_HCC135",
            "HHS_HCC138",
            "HHS_HCC145",
            "HHS_HCC146",
            "HHS_HCC154",
            "HHS_HCC156",
            "HHS_HCC163",
            "HHS_HCC187",
            "HHS_HCC253",
        ]

        severity_3_hccs = [
            "HHS_HCC001",
            "HHS_HCC003",
            "HHS_HCC006",
            "HHS_HCC010",
            "HHS_HCC011",
            "HHS_HCC012",
            "HHS_HCC027",
            "HHS_HCC045",
            "HHS_HCC054",
            "HHS_HCC055",
            "HHS_HCC061",
            "HHS_HCC063",
            "HHS_HCC066",
            "HHS_HCC074",
            "HHS_HCC075",
            "HHS_HCC081",
            "HHS_HCC082",
            "HHS_HCC083",
            "HHS_HCC084",
            "HHS_HCC096",
            "HHS_HCC108",
            "HHS_HCC109",
            "HHS_HCC110",
            "HHS_HCC113",
            "HHS_HCC114",
            "HHS_HCC117",
            "HHS_HCC119",
            "HHS_HCC121",
            "HHS_HCC132",
            "HHS_HCC139",
            "HHS_HCC142",
            "HHS_HCC149",
            "HHS_HCC150",
            "HHS_HCC159",
            "HHS_HCC218",
            "HHS_HCC223",
            "HHS_HCC226",
            "HHS_HCC228",
        ]

        severity_2_hccs = [
            "HHS_HCC004",
            "HHS_HCC013",
            "HHS_HCC019",
            "HHS_HCC020",
            "HHS_HCC021",
            "HHS_HCC023",
            "HHS_HCC028",
            "HHS_HCC029",
            "HHS_HCC036",
            "HHS_HCC046",
            "HHS_HCC047",
            "HHS_HCC048",
            "HHS_HCC056",
            "HHS_HCC057",
            "HHS_HCC062",
            "HHS_HCC069",
            "HHS_HCC070",
            "HHS_HCC097",
            "HHS_HCC120",
            "HHS_HCC151",
            "HHS_HCC153",
            "HHS_HCC160",
            "HHS_HCC161_1",
            "HHS_HCC162",
            "HHS_HCC188",
            "HHS_HCC217",
            "HHS_HCC219",
        ]

        severity_1_hccs = [
            "HHS_HCC037_1",
            "HHS_HCC037_2",
            "HHS_HCC071",
            "HHS_HCC102",
            "HHS_HCC103",
            "HHS_HCC118",
            "HHS_HCC161_2",
            "HHS_HCC234",
            "HHS_HCC254",
        ]

        # This is the default Severity Status
        severity_status = "Severity1"

        if any(category in category_list for category in severity_5_hccs):
            severity_status = "Severity5"
        elif any(category in category_list for category in severity_4_hccs):
            severity_status = "Severity4"
        elif any(category in category_list for category in severity_3_hccs):
            severity_status = "Severity3"
        elif any(category in category_list for category in severity_2_hccs):
            severity_status = "Severity2"
        elif any(category in category_list for category in severity_1_hccs):
            severity_status = "Severity1"

        return severity_status

    # --- Helper methods which should not be overwritten ---

    def _apply_csr_adj(self, score: float) -> float:
        """
        Applies normalization factor and coding intensity adjustment to the score.
        Per CMS documentation, rounding happens at each step. Technically, it should
        be to the third decmial, but rounding to the 4th for greater precision.

        Args:
            score (float): The score to be adjusted.

        Returns:
            float: The adjusted score.
        """
        return round(score * self.csr_adjuster, 3)



from typing import Union


class GenericCodeCategory:
    """
    Encapsulates a generic code and its corresponding category mapping. This is a base
    class and should not be called directly. Some codes can go to multiple categories,
    thus the categories attribute is a list.

    Attributes:
        category_map (dict): A dictionary containing the mapping of codes to categories.
        mapper_code (str): The code to be mapped.
        type (str): The type of code (default is None).
        categories (list[str]): A list containing the categories corresponding to the code.
    """

    def __init__(self, category_map: dict, code: str, type: Union[str, None] = None):
        """
        Initializes GenericCodeCategory with the provided parameters.

        Args:
            category_map (dict): A dictionary containing the mapping of codes to categories.
            code (str): The code to be mapped.
            type (str, optional): The type of code (default is None).
        """
        self.mapper_code = code
        self.type = type
        self.category_map = category_map[type]
        self.categories = self.category_map.get(code, [None])


class DxCodeCategory(GenericCodeCategory):
    """
    Represents a diagnosis code and its category mapping.

    Attributes:
        category_map (dict): A dictionary containing the mapping of codes to categories.
        code (str): The diagnosis code.
        type (str): The type of code, defaulted to "diag".
    """

    def __init__(self, category_map: dict, code: str, type: str = "diag"):
        """
        Initializes DxCodeCategory with the provided parameters.

        Args:
            category_map (dict): A dictionary containing the mapping of codes to categories.
            code (str): The diagnosis code.
            type (str, optional): The type of code (default is "diag").
        """
        super().__init__(category_map, code, type)


class NDCCodeCategory(GenericCodeCategory):
    """
    Represents a National Drug Code (NDC) and its category mapping.

    Attributes:
        category_map (dict): A dictionary containing the mapping of codes to categories.
        code (str): The NDC code.
        type (str): The type of code, defaulted to "ndc".
    """

    def __init__(self, category_map: dict, code: str, type: str = "ndc"):
        """
        Initializes NDCCodeCategory with the provided parameters.

        Args:
            category_map (dict): A dictionary containing the mapping of codes to categories.
            code (str): The NDC code.
            type (str, optional): The type of code (default is "ndc").
        """
        super().__init__(category_map, code, type)


class ProcCodeCategory(GenericCodeCategory):
    """
    Represents a procedure code (CPT, HCPCS, ICD10-CM codes, etc.) and its category mapping.

    Attributes:
        category_map (dict): A dictionary containing the mapping of codes to categories.
        code (str): The procedure code.
        type (str): The type of code, defaulted to "proc".
    """

    def __init__(self, category_map: dict, code: str, type: str = "proc"):
        """
        Initializes ProcCodeCategory with the provided parameters.

        Args:
            category_map (dict): A dictionary containing the mapping of codes to categories.
            code (str): The procedure code.
            type (str, optional): The type of code (default is "proc").
        """
        super().__init__(category_map, code, type)



from typing import Union, Type, List
from .utilities import determine_age_band
from .beneficiary import MedicareBeneficiary
from .category import Category
from .result import MedicareScoringResult
from .mapper import DxCodeCategory
from .model import BaseModel


class MedicareModel(BaseModel):
    """
    This is the foundation class for Medicare Models. It is not to be called directly. It loads all
    relevant information for that model and year as class attributes.

    How this class works:
    1.  Instantiate the class with set up information: model version and year (optional).
        If year is None, the code will determine the most recent year and use that.
    2.  Instantiating the class loads all reference information into memory for performance purposes.
    3.  The entry point is the "score" method.
    4.  Since each model has its own nuances, each model is its own class inheriting from this class.

    Attributes:
        coding_intensity_adjuster (float): CMS is required by law to adjust Medicare Advantage risk scores
                                           using this value. See:
                                           https://bettermedicarealliance.org/wp-content/uploads/2022/06/BMA-Fact-Sheet-Coding-Practices-and-Adjustments-in-Medicare-Advantage-1.pdf
        normalization_factor (float): CMS updates normalization factor each year to apply to the risk scores. See:
                                      https://www.commonwealthfund.org/publications/explainer/2024/mar/how-government-updates-payment-rates-medicare-advantage-plans

    Notes:
        In terms of inheritance for child classes and how it relates to methods, the code is
        organizaed in the following fashion:

        Methods strongly advised against overwriting:
            score
            _build_category_details
        Methods unlikely needing overwriting but could happen based on needs:
            _apply_hierarchies
            _determine_demographic_categories
            _get_dx_categories
            _get_coding_intensity_adjuster
        Methods likely needing overwriting, e.g.:
            _determine_age_gender_category
            _determine_disease_interactions
            _age_sex_edits
            _get_normalization_factor
        Helper methods to not override, e.g.: _apply_norm_factor_coding_adj
    """

    def __init__(self, version: str, year: Union[int, None] = None):
        super().__init__(lob="medicare", version=version, year=year)
        self.coding_intensity_adjuster = self._get_coding_intensity_adjuster(
            self.model_year
        )
        self.normalization_factor = self._get_normalization_factor(self.model_year)

    def score(
        self,
        gender: str,
        orec: str,
        medicaid: bool,
        diagnosis_codes: Union[List[str], None] = None,
        age: Union[int, None] = None,
        dob: Union[str, None] = None,
        population: str = "CNA",
        verbose: bool = False,
    ) -> Type[MedicareScoringResult]:
        """
        Determines the risk score for the inputs. Entry point for end users.

        Steps:
        1. Use beneficiary information to get the demographic categories
        2. Using diagnosis code inputs and beneficiary information get the diagnosis code to
           category relationship
        3. Get the unique set of categories from diagnosis codes
        4. Apply hierarchies
        5. Determine disease interactions

        Args:
            gender (str): Gender of the beneficiary being scored, valid values M or F.
            orec (str): Original Entitlement Reason Code of the beneficiary. See: https://bluebutton.cms.gov/assets/ig/ValueSet-orec.html for valid values
            medicaid (bool): Beneficiary medicaid status, True or False
            diagnosis_codes (list): List of the diagnosis codes associated with the beneficiary
            age (int): Age of the beneficiary, can be None.
            dob (str): Date of birth of the beneficiary, can be None
            population (str): Population of beneficiary being scored, valid values are CNA, CND, CPA, CPD, CFA, CFD, INS, NE
            verbose (bool): Indicates if trimmed output or full output is desired

        Returns:
            MedicareScoringResult: An instantiated object of ScoringResult class.
        """
        beneficiary = MedicareBeneficiary(
            gender, orec, medicaid, population, age, dob, self.model_year
        )
        demo_categories = self._determine_demographic_categories(beneficiary)

        if diagnosis_codes:
            cat_dict = {}
            dx_categories = self._get_dx_categories(diagnosis_codes, beneficiary)
            # Some diagnosis codes go to more than one category thus the category is a list
            # and it is a two step process to unpack them
            unique_disease_cats = set(
                category
                for dx_code in dx_categories
                for category in dx_code.categories
                if category is not None and category != "NA"
            )

            for category in unique_disease_cats:
                # This is done to obtain a consistent output for diagnosis map
                # If there are no diagnosis codes mapping to the category it should
                # return None, as opposed to an empty list. The below statement making
                # the map would create an empty list, thus the if statement following
                # to catch that and modify it
                diagnosis_map = [
                    dx_code.mapper_code
                    for dx_code in dx_categories
                    if category in dx_code.categories
                ]
                cat_dict[category] = diagnosis_map
        else:
            cat_dict = {}
            unique_disease_cats = None

        if unique_disease_cats:
            unique_categories = demo_categories + list(unique_disease_cats)
        else:
            unique_categories = demo_categories

        categories = [
            Category(
                self.reference_files,
                beneficiary.risk_model_population,
                category,
                cat_dict.get(category),
            )
            for category in unique_categories
        ]
        categories = self._apply_hierarchies(categories)
        categories = self._determine_disease_interactions(categories, beneficiary)

        score_raw = sum([category.coefficient for category in categories])
        disease_score_raw = sum(
            [
                category.coefficient
                for category in categories
                if "disease" in category.type
            ]
        )
        demographic_score_raw = sum(
            [
                category.coefficient
                for category in categories
                if "demographic" in category.type
            ]
        )

        category_details = self._build_category_details(categories, verbose)

        results = MedicareScoringResult(
            gender=beneficiary.gender,
            orec=beneficiary.orec,
            medicaid=beneficiary.medicaid,
            age=beneficiary.age,
            dob=beneficiary.dob,
            diagnosis_codes=diagnosis_codes,
            year=self.year,
            population=beneficiary.population,
            risk_model_age=beneficiary.risk_model_age,
            risk_model_population=beneficiary.risk_model_population,
            model_version=self.version,
            model_year=self.model_year,
            coding_intensity_adjuster=self.coding_intensity_adjuster,
            normalization_factor=self.normalization_factor,
            score_raw=score_raw,
            disease_score_raw=disease_score_raw,
            demographic_score_raw=demographic_score_raw,
            score=self._apply_norm_factor_coding_adj(score_raw),
            disease_score=self._apply_norm_factor_coding_adj(disease_score_raw),
            demographic_score=self._apply_norm_factor_coding_adj(demographic_score_raw),
            category_list=[category.category for category in categories],
            category_details=category_details,
        )

        return results

    def _build_category_details(
        self, categories: List[Type[Category]], verbose: bool
    ) -> dict:
        # Combine the dictionaries to make output
        category_details = {}
        for category in categories:
            if verbose:
                category_details[category.category] = {
                    "coefficient": category.coefficient,
                    "type": category.type,
                    "category_number": category.number,
                    "category_description": category.description,
                    "dropped_categories": category.dropped_categories,
                    "diagnosis_map": category.mapper_codes,
                }
            else:
                category_details[category.category] = {
                    "coefficient": category.coefficient,
                    "diagnosis_map": category.mapper_codes,
                }

        return category_details

    # --- Methods which may need to be overwritten but unlikely to be overwritten ---

    def _determine_demographic_categories(
        self, beneficiary: Type[MedicareBeneficiary]
    ) -> List[str]:
        """
        Determine demographic categories based on beneficiary attributes.

        Args:
            beneficiary (Type[MedicareBeneficiary]): An instance of MedicareBeneficiary.

        Returns:
            list: A list containing demographic categories.
        """
        demo_cats = []
        demo_cats.append(
            self._determine_age_gender_category(
                beneficiary.age, beneficiary.gender, beneficiary.population
            )
        )
        demo_int = self._determine_demographic_interactions(
            beneficiary.gender, beneficiary.orig_disabled, beneficiary.medicaid
        )
        if demo_int:
            demo_cats.extend(demo_int)

        return demo_cats

    def _apply_hierarchies(
        self, categories: List[Type[Category]]
    ) -> List[Type[Category]]:
        """
        Filters out categories falling into hierarchies per the model hierarchy_definition file.

        Args:
            categories (List[Type[Category]]): List of category objects to process.

        Returns:
            List[Type[Category]]: List of category objects after filtering.

        Notes:
            For each category, the codes dropped are tracked and assigned to the
            attribute "dropped_categories" of the category object.
        """
        category_list = [category.category for category in categories]
        dropped_codes_total = []

        for category in categories:
            dropped_codes = []
            if category.category in self.reference_files.hierarchy_definitions.keys():
                for remove_category in self.reference_files.hierarchy_definitions[
                    category.category
                ]["remove_code"]:
                    if remove_category in category_list:
                        dropped_codes.append(remove_category)
                        dropped_codes_total.append(remove_category)

            if dropped_codes:
                category.dropped_categories = dropped_codes

        # Remove objects from list
        final_categories = [
            category
            for category in categories
            if category.category not in dropped_codes_total
        ]

        return final_categories

    def _get_dx_categories(
        self, diagnosis_codes: List[str], beneficiary: Type[MedicareBeneficiary]
    ) -> List[Type[DxCodeCategory]]:
        """
        Generates diagnosis code to categories relationships based on provided diagnosis codes
        and beneficiary information.

        Args:
            diagnosis_codes (List[str]): List of diagnosis codes.
            beneficiary (Type[MedicareBeneficiary]): Instance of MedicareBeneficiary representing the beneficiary information.

        Returns:
            List[Type[DxCodeCategory]]: List of DxCodeCategory objects representing the diagnosis code categories.
        """
        dx_categories = [
            DxCodeCategory(self.reference_files.category_map, diagnosis_code)
            for diagnosis_code in diagnosis_codes
        ]

        for dx in dx_categories:
            edit_category = self._age_sex_edits(
                beneficiary.gender, beneficiary.age, dx.mapper_code
            )
            if edit_category:
                dx.categories = edit_category

        return dx_categories

    def _get_coding_intensity_adjuster(self, year: int) -> float:
        """
        Gets the appropriate coding intensity adjuster for the model year as outlined by
        CMS in their annual Annoucements:
        https://www.cms.gov/medicare/payment/medicare-advantage-rates-statistics/announcements-and-documents

        Returns:
            float: The coding intensity adjuster.

        Notes:
            In the documents, the coding intensity adjuster is listed as a small decimal, e.g.
            .059. In scoring calculations, the adjuster is applied by doing score * (1-.059).
            Thus, the below already represents the 1-.059 for convenience.
            Most years, the coding intensity adjuster is the statuatory minimum of .059.
        """
        coding_intensity_dict = {
            2020: 0.941,
            2021: 0.941,
            2022: 0.941,
            2023: 0.941,
            2024: 0.941,
            2025: 0.941,
        }
        if coding_intensity_dict.get(year):
            coding_intensity_adjuster = coding_intensity_dict.get(year)
        else:
            coding_intensity_adjuster = 1

        return coding_intensity_adjuster

    # --- Methods likely to be overwritten by each model class ---

    def _determine_age_gender_category(
        self, age: int, gender: str, population: str
    ) -> str:
        """
        Determines the demographic category based on age, gender, and population.

        Args:
            age (int): Age of the individual.
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            population (str): Beneficiary model population used for scoring

        Returns:
            str: Demographic category based on age, gender, and population.
        """
        if population[:2] == "NE":
            demo_category_ranges = [
                "0_34",
                "35_44",
                "45_54",
                "55_59",
                "60_64",
                "65",
                "66",
                "67",
                "68",
                "69",
                "70_74",
                "75_79",
                "80_84",
                "85_89",
                "90_94",
                "95_GT",
            ]
        else:
            demo_category_ranges = [
                "0_34",
                "35_44",
                "45_54",
                "55_59",
                "60_64",
                "65_69",
                "70_74",
                "75_79",
                "80_84",
                "85_89",
                "90_94",
                "95_GT",
            ]

        demographic_category_range = determine_age_band(age, demo_category_ranges)

        if population[:2] == "NE":
            demographic_category = f"NE{gender}{demographic_category_range}"
        else:
            demographic_category = f"{gender}{demographic_category_range}"

        return demographic_category

    def _determine_demographic_interactions(
        self, gender: str, orig_disabled: bool, medicaid: bool
    ) -> List[str]:
        """
        Determines demographic interactions based on gender, disability status, and Medicaid enrollment.

        Args:
            gender (str): Gender of the individual ('M' for male, 'F' for female).
            orig_disabled (bool): Indicates if the individual was originally disabled.
            medicaid (bool): Indicates if the individual is enrolled in Medicaid.

        Returns:
            List[str]: List of demographic interaction labels.
        """
        demo_interactions = []
        if gender == "F" and orig_disabled:
            demo_interactions.append("OriginallyDisabled_Female")
        elif gender == "M" and orig_disabled:
            demo_interactions.append("OriginallyDisabled_Male")

        if medicaid:
            demo_interactions.append("LTIMCAID")

        return demo_interactions

    def _determine_disease_interactions(
        self, categories: List[Type[Category]], beneficiary: Type[MedicareBeneficiary]
    ) -> List[Type[Category]]:
        """
        Determines disease interactions based on provided Category objects and beneficiary information.
        Placeholder to be overwritten by child classes.

        Args:
            categories (List[Type[Category]]): List of Category objects representing disease categories.
            beneficiary (Type[MedicareBeneficiary]): Instance of MedicareBeneficiary representing the beneficiary information.

        Returns:
            List[Type[Category]]: List of Category objects representing the disease interactions.
        """
        interaction_list = []
        category_count = self._determine_payment_count_category(categories)
        if category_count:
            interaction_list.append(category_count)

        interactions = [
            Category(self.reference_files, beneficiary.risk_model_population, category)
            for category in interaction_list
        ]

        interactions.append(categories)

        return interactions

    def _determine_payment_count_category(self, categories: list) -> str:
        """
        Determines the payment count category based on the number of categories provided.

        Args:
            categories (list): List of categories.

        Returns:
            str: Payment count category.
        """
        category_count = len(categories)
        category = None
        if category_count > 9:
            category = "D10P"
        elif category_count > 0:
            category = f"D{category_count}"

        return category

    def _age_sex_edits(
        self, gender: str, age: int, diagnosis_code: str
    ) -> Union[List[str], None]:
        """
        Placeholder function to be overwritten by child clasess. This to encapsulate
        the age sex edits for a model that are to be performed on the
        DxCodeCategory objects in the _get_dx_categories method.

        Returns:
            List[str]: List of categories based on input gender, age, diagnosis code
        """
        return ["NA"]

    def _get_normalization_factor(self, year: int) -> float:
        """
        Placeholder function to be overwritten by child class to return the normalization
        factor for model and year combination.

        Returns:
            float: The normalization factor.
        """
        normalization_factor = 1

        return normalization_factor

    # --- Helper methods which should not be overwritten ---

    def _apply_norm_factor_coding_adj(self, score: float) -> float:
        """
        Applies normalization factor and coding intensity adjustment to the score.
        Per CMS documentation, rounding happens at each step. Technically, it should
        be to the third decmial, but rounding to the 4th for greater precision.

        Args:
            score (float): The score to be adjusted.

        Returns:
            float: The adjusted score.
        """
        return round(
            round(score * self.coding_intensity_adjuster, 4)
            / self.normalization_factor,
            4,
        )
